"""Test for the spickle module.

"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2007, Enthought, Inc.
# License: BSD Style.


import unittest
import numpy
from pickle import dumps

try:
    from apptools.persistence import spickle
except ImportError:
    import nose
    raise nose.SkipTest('spickle is not supported with Python3')

from traits.api import HasTraits, Float, Int

class A:
    def __init__(self):
        self.a = 100
        self.array = numpy.linspace(0, 1, 5)

class B(HasTraits):
    i = Int(10)
    f = Float(1.0)

class Foo(object):
    def __init__(self, a=1):
        self.a = A()
        self.a.b = 200
        self.ref = self.a
        self.b = B()
        self.b.set(i=20, f=2.0)

class TestStatePickler(unittest.TestCase):
    def _test_object(self, x):
        assert x.a.a == 100
        assert numpy.all(x.a.array == numpy.linspace(0, 1, 5))
        assert x.a.b == 200
        assert x.a == x.ref
        assert x.b.i == 20
        assert x.b.f == 2.0

    def test_dump_state(self):
        "Test if we are able to dump the state to a string."
        f = Foo()
        str = dumps(f)
        st = spickle.get_state(f)
        str1 = spickle.dumps_state(st)
        self.assertEqual(str, str1)
        st = spickle.loads_state(str)
        self.assertEqual(str, spickle.dumps_state(st))

    def test_state2object(self):
        "Test if we can convert a state to an object."
        f = Foo()
        str = dumps(f)
        st = spickle.get_state(f)
        g = spickle.state2object(st)
        self._test_object(g)


if __name__ == "__main__":
    unittest.main()



#-------------------------------------------------------------------------------
#
#  Defines the TemplateDataNames class used to manage application data source
#  bindings to named context data.
#
#  Written by: David C. Morrill
#
#  Date: 07/26/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the TemplateDataNames class used to manage application data source
    bindings to named context data.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Instance, Int, List, Property, Delegate, \
           cached_property

from traitsui.api \
    import View, Item, TableEditor, EnumEditor, TextEditor

from traitsui.table_column \
    import ObjectColumn

from .template_data_name \
    import TemplateDataName

from .itemplate_data_context \
    import ITemplateDataContext

from .template_choice \
    import TemplateChoice

#-------------------------------------------------------------------------------
#  Table editor support for editing a list of TemplateDataName objects:
#-------------------------------------------------------------------------------

class BindingsColumn ( ObjectColumn ):

    # The column index:
    index = Int

    #-- ObjectColumn Method Overrides ------------------------------------------

    def is_editable ( self, object ):
        return (self.index < len( object.data_name.items ))

    def get_editor ( self, object ):
        return EnumEditor( values =
             [ dnic.choice_value for dnic in object.data_name.items[
                                         self.index ].data_name_item_choices ] +
             [ self.get_raw_value( object ) ] )

class ResolvedColumn ( ObjectColumn ):

    def get_raw_value( self, object ):
        if object.data_name.resolved:
            return ''

        return 'X'

    def get_cell_color ( self, object ):
        if object.data_name.resolved:
            return self.read_only_cell_color_

        return self.cell_color_

class OptionalColumn ( ObjectColumn ):

    def get_raw_value( self, object ):
        return ' X'[ object.data_name.optional ]

# Define the table editor:
table_editor = TableEditor(
    columns_name       = 'table_columns',
    configurable       = False,
    auto_size          = False,
    sortable           = False,
    scroll_dy          = 4,
    selection_bg_color = None
)

# The standard columns:
std_columns = [
    ResolvedColumn( name     = 'resolved',
                    label    = '?',
                    editable = False,
                    width    = 20,
                    horizontal_alignment = 'center',
                    cell_color = 0xFF8080 ),
    OptionalColumn( name     = 'optional',
                    label    = '*',
                    editable = False,
                    width    = 20,
                    horizontal_alignment = 'center' ),
    ObjectColumn(   name     = 'description',
                    editor   = TextEditor(),
                    width    = 0.47 ) ]

#-------------------------------------------------------------------------------
#  'TemplateDataNames' class:
#-------------------------------------------------------------------------------

class TemplateDataNames ( HasPrivateTraits ):

    #-- Public Traits ----------------------------------------------------------

    # The data context to which bindings are made:
    context = Instance( ITemplateDataContext )

    # The current set of data names to be bound to the context:
    data_names = List( TemplateDataName )

    # The list of unresolved, required bindings:
    unresolved_data_names = Property( depends_on = 'data_names.resolved' )

    # The list of optional bindings:
    optional_data_names = Property( depends_on = 'data_names.optional' )

    # The list of unresolved optional bindings:
    unresolved_optional_data_names = Property(
        depends_on = 'data_names.[resolved,optional]' )

    #-- Private Traits ---------------------------------------------------------

    # List of 'virtual' data names for use by table editor:
    virtual_data_names = List

    # The list of table editor columns:
    table_columns = Property( depends_on = 'data_names' ) # List( ObjectColumn )

    #-- Traits View Definitions ------------------------------------------------

    view = View(
        Item( 'virtual_data_names',
              show_label = False,
              style      = 'custom',
              editor     = table_editor )
    )

    #-- Property Implementations -----------------------------------------------

    @cached_property
    def _get_unresolved_data_names ( self ):
        return [ dn for dn in self.data_names
                 if (not dn.resolved) and (not dn.optional) ]

    @cached_property
    def _get_optional_data_names ( self ):
        return [ dn for dn in self.data_names if dn.optional ]

    @cached_property
    def _get_unresolved_optional_data_names ( self ):
        return [ dn for dn in self.data_names
                 if (not dn.resolved) and dn.optional ]

    @cached_property
    def _get_table_columns ( self ):
        n =  max( [ len( dn.items ) for dn in self.data_names ] )
        if n == 1:
            return std_columns + [ BindingsColumn( name  = 'value0',
                                                   label = 'Name',
                                                   width = 0.43 ) ]
        width = 0.43 / n
        return (std_columns +
                [ BindingsColumn( name = 'value%d' % i,
                                 index = i,
                                 label = 'Name %d' % ( i + 1 ),
                                 width = width ) for i in range( n ) ])

    #-- Trait Event Handlers ---------------------------------------------------

    def _context_changed ( self, context ):
        for data_name in self.data_names:
            data_name.context = context

    def _data_names_changed ( self, old, new ):
        """ Handles the list of 'data_names' being changed.
        """
        # Make sure that all of the names are unique:
        new = set( new )

        # Update the old and new context links:
        self._update_contexts( old, new )

        # Update the list of virtual names based on the new set:
        dns = [ VirtualDataName( data_name = dn ) for dn in new ]
        dns.sort( lambda l, r: cmp( l.description, r.description ) )
        self.virtual_data_names = dns

    def _data_names_items_changed ( self, event ):
        # Update the old and new context links:
        old, new = event.old, event.new
        self._update_contexts( old, new )

        # Update the list of virtual names based on the old and new sets:
        i = event.index
        self.virtual_data_names[ i: i + len( old ) ] = [
            VirtualDataName( data_name = dn ) for dn in new ]

    #-- Private Methods --------------------------------------------------------

    def _update_contexts ( self, old, new ):
        """ Updates the data context for an old and new set of data names.
        """
        for data_name in old:
            data_name.context = None

        context = self.context
        for data_name in new:
            data_name.context = context

#-------------------------------------------------------------------------------
#  'VirtualDataName' class:
#-------------------------------------------------------------------------------

# Define the 'VirtualValue' property:
def _get_virtual_data ( self, name ):
    return self.data_name.items[
               self.trait( name ).index ].data_name_item_choice.choice_value

def _set_virtual_data ( self, name, new_value ):
    old_value = _get_virtual_data( self, name )
    if old_value != new_value:
        self.data_name.items[
                 self.trait( name ).index ].data_name_item_choice = \
            TemplateChoice( choice_value = new_value )

        self.trait_property_changed( name, old_value, new_value )

VirtualValue = Property( _get_virtual_data, _set_virtual_data )

class VirtualDataName ( HasPrivateTraits ):

    # The TemplateDataName this is a virtual copy of:
    data_name = Instance( TemplateDataName )

    # The data name description:
    description = Delegate( 'data_name', modify = True )

    # The 'virtual' traits of this object:
    value0 = VirtualValue( index = 0 )
    value1 = VirtualValue( index = 1 )
    value2 = VirtualValue( index = 2 )
    value3 = VirtualValue( index = 3 )
    value4 = VirtualValue( index = 4 )
    value5 = VirtualValue( index = 5 )




""" A generic object factory. """


# Enthought library imports.
from traits.api import Any

# Local imports.
from .abstract_factory import AbstractFactory


class Factory(AbstractFactory):
    """ A generic object factory.

    This implementation of the abstract factory interface provides for the
    common scenario where the factory produces objects of exactly one type.

    """

    #### 'Factory' interface ##################################################

    # The type of object that we create.
    #
    # fixme: This trait definition should be 'Class' but currently this only
    # allows old-style classes!
    target_class = Any

    ###########################################################################
    # 'AbstractFactory' interface.
    ###########################################################################

    def can_create(self, target_class, *args, **kw):
        """ Returns True if the factory can create objects of a class. """

        return target_class is self.target_class

    def create(self, *args, **kw):
        """ Creates an object! """

        return self.target_class(*args, **kw)

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Interface, List, Str


class UserStorageError(Exception):
    """This is the exception raised by an IUserStorage object when an error
    occurs accessing the database.  Its string representation is displayed as
    an error message to the user."""


class IUserStorage(Interface):
    """This defines the interface expected by a UserManager instance to handle
    the low level storage of the user data."""

    #### 'IUserStorage' interface #############################################

    # A list of strings describing the storage capabilities.  'user_password'
    # means a user's password can be changed.  'user_add' means a user can be
    # added.  'user_modify' means a user can be modified.  'user_delete' means
    # a user can be deleted.
    capabilities = List(Str)

    ###########################################################################
    # 'IUserStorage' interface.
    ###########################################################################

    def add_user(self, name, description, password):
        """Add a new user with the given name, description and password."""

    def authenticate_user(self, name, password):
        """Return a tuple of the name, description, and blob of the user with
        the given name if they are successfully authenticated with the given
        password.  A tuple of empty strings will be returned if the
        authentication failed."""

    def delete_user(self, name):
        """Delete the user with the given name (which will not be empty)."""

    def is_empty(self):
        """Return True if the user database is empty.  It will only ever be
        called once."""

    def matching_users(self, name):
        """Return a list of tuples of the full name and description of all
        users, sorted by the full name, that match the given name.  How the
        name is interpreted (eg. as a regular expression) is determined by the
        storage."""

    def modify_user(self, name, description, password):
        """Update the description and password for the user with the given name
        (which will not be empty)."""

    def unauthenticate_user(self, user):
        """Unauthenticate the given user (which is an object implementing
        IUser) and return True if it was successful."""

    def update_blob(self, name, blob):
        """Update the blob for the user with the given name (which will not be
        empty)."""

    def update_password(self, name, password):
        """Update the password for the user with the given name (which will not
        be empty)."""



""" Tests for scoped preferences. """


# Standard library imports.
import os, tempfile, unittest
from os.path import join

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from apptools.preferences.api import Preferences, ScopedPreferences

# Local imports.
from .preferences_test_case import PreferencesTestCase


# This module's package.
PKG = 'apptools.preferences.tests'


class ScopedPreferencesTestCase(PreferencesTestCase):
    """ Tests for the scoped preferences. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        self.preferences = ScopedPreferences()

        # The filename of the example preferences file.
        self.example = resource_filename(PKG, 'example.ini')

        # A temporary directory that can safely be written to.
        self.tmpdir = tempfile.mkdtemp()

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        # Remove the temporary directory.
        os.removedirs(self.tmpdir)

        return

    ###########################################################################
    # Tests overridden from 'PreferencesTestCase'.
    ###########################################################################

    def test_node(self):
        """ node """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node('acme')
        self.assertNotEqual(None, node)
        self.assertEqual('acme', node.name)
        self.assertEqual('acme', node.path)
        self.assertEqual(p.node('application/'), node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node('acme'))

        # Try a nested path.
        node = p.node('acme.ui')
        self.assertNotEqual(None, node)
        self.assertEqual('ui', node.name)
        self.assertEqual('acme.ui', node.path)
        self.assertEqual(p.node('application/acme'), node.parent)

        # And just to be sure, a really nested path.
        node = p.node('acme.ui.splash_screen')
        self.assertNotEqual(None, node)
        self.assertEqual('splash_screen', node.name)
        self.assertEqual('acme.ui.splash_screen', node.path)
        self.assertEqual(p.node('application/acme.ui'), node.parent)

        return

    def test_save(self):
        """ save """

        p = self.preferences

        # Get the application scope.
        application = p.node('application/')

        tmp = join(self.tmpdir, 'test.ini')
        application.filename = tmp

        # Set a value.
        p.set('acme.ui.bgcolor', 'red')

        # Save all scopes.
        p.save()

        # Make sure a file was written.
        self.assertEqual(True, os.path.exists(tmp))

        # Load the 'ini' file into a new preferences node and make sure the
        # preference is in there.
        p = Preferences()
        p.load(tmp)

        self.assertEqual('red', p.get('acme.ui.bgcolor'))

        # Cleanup.
        os.remove(tmp)

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_ability_to_specify_primary_scope(self):

        preferences = ScopedPreferences(
            scopes = [
                Preferences(name='a'),
                Preferences(name='b'),
                Preferences(name='c')
            ],

            primary_scope_name = 'b'
        )

        # This should set the prefrrence in the primary scope.
        preferences.set('acme.foo', 'bar')

        # Look it up specifically in the primary scope.
        self.assertEqual('bar', preferences.get('b/acme.foo'))

        return

    def test_builtin_scopes(self):
        """ builtin scopes """

        p = self.preferences

        # Make sure the default built-in scopes get created.
        self.assertEqual(True, p.node_exists('application/'))
        self.assertEqual(True, p.node_exists('default/'))

        return

    def test_get_and_set_in_specific_scope(self):
        """ get and set in specific scope """

        p = self.preferences

        # Set a preference and make sure we can get it again!
        p.set('default/acme.ui.bgcolor', 'red')
        self.assertEqual('red', p.get('default/acme.ui.bgcolor'))

        return

    def test_clear_in_specific_scope(self):
        """ clear in specific scope """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set('application/acme.ui.bgcolor', 'red')
        p.set('default/acme.ui.bgcolor', 'yellow')

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual('red', p.get('acme.ui.bgcolor'))

        # Now clear out the application scope.
        p.clear('application/acme.ui')
        self.assertEqual(0, len(p.keys('application/acme.ui')))

        # We should now get the value from the default scope.
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))

        return

    def test_remove_in_specific_scope(self):
        """ remove in specific scope """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set('application/acme.ui.bgcolor', 'red')
        p.set('default/acme.ui.bgcolor', 'yellow')

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual('red', p.get('acme.ui.bgcolor'))

        # Now remove it from the application scope.
        p.remove('application/acme.ui.bgcolor')

        # We should now get the value from the default scope.
        self.assertEqual('yellow', p.get('acme.ui.bgcolor'))

        return

    def test_keys_in_specific_scope(self):
        """ keys in specific scope """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.keys('default/'))

        # Set some preferences in the node.
        p.set('default/a', '1')
        p.set('default/b', '2')
        p.set('default/c', '3')

        keys = p.keys('default/')
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        # Set some preferences in a child node.
        p.set('default/acme.a', '1')
        p.set('default/acme.b', '2')
        p.set('default/acme.c', '3')

        keys = p.keys('default/acme')
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        # And, just to be sure, in a child of the child node ;^)
        p.set('default/acme.ui.a', '1')
        p.set('default/acme.ui.b', '2')
        p.set('default/acme.ui.c', '3')

        keys = p.keys('default/acme.ui')
        keys.sort()

        self.assertEqual(['a', 'b', 'c'], keys)

        return

    def test_node_in_specific_scope(self):
        """ node in specific scope """

        p = self.preferences

        # Try an empty path.
        self.assertEqual(p, p.node())

        # Try a simple path.
        node = p.node('default/acme')
        self.assertNotEqual(None, node)
        self.assertEqual('acme', node.name)
        self.assertEqual('acme', node.path)
        self.assertEqual(p.node('default/'), node.parent)

        # Make sure we get the same node each time we ask for it!
        self.assertEqual(node, p.node('default/acme'))

        # Try a nested path.
        node = p.node('default/acme.ui')
        self.assertNotEqual(None, node)
        self.assertEqual('ui', node.name)
        self.assertEqual('acme.ui', node.path)
        self.assertEqual(p.node('default/acme'), node.parent)

        # And just to be sure, a really nested path.
        node = p.node('default/acme.ui.splash_screen')
        self.assertNotEqual(None, node)
        self.assertEqual('splash_screen', node.name)
        self.assertEqual('acme.ui.splash_screen', node.path)
        self.assertEqual(p.node('default/acme.ui'), node.parent)

        return

    def test_node_exists_in_specific_scope(self):
        """ node exists """

        p = self.preferences

        self.assertEqual(True, p.node_exists())
        self.assertEqual(False, p.node_exists('default/acme'))

        p.node('default/acme')
        self.assertEqual(True, p.node_exists('default/acme'))

        return

    def test_node_names_in_specific_scope(self):
        """ node names in specific scope """

        p = self.preferences

        # It should be empty to start with!
        self.assertEqual([], p.node_names('default/'))

        # Create some nodes.
        p.node('default/a')
        p.node('default/b')
        p.node('default/c')

        names = p.node_names('default/')
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        # Creatd some nodes in a child node.
        p.node('default/acme.a')
        p.node('default/acme.b')
        p.node('default/acme.c')

        names = p.node_names('default/acme')
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        # And, just to be sure, in a child of the child node ;^)
        p.node('default/acme.ui.a')
        p.node('default/acme.ui.b')
        p.node('default/acme.ui.c')

        names = p.node_names('default/acme.ui')
        names.sort()

        self.assertEqual(['a', 'b', 'c'], names)

        return

    def test_default_lookup_order(self):
        """ default lookup order """

        p = self.preferences

        # Set a value in both the application and default scopes.
        p.set('application/acme.ui.bgcolor', 'red')
        p.set('default/acme.ui.bgcolor', 'yellow')

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual('red', p.get('acme.ui.bgcolor'))

        # But we can still get at each scope individually.
        self.assertEqual('red', p.get('application/acme.ui.bgcolor'))
        self.assertEqual('yellow', p.get('default/acme.ui.bgcolor'))

        return

    def test_lookup_order(self):
        """ lookup order """

        p = self.preferences
        p.lookup_order = ['default', 'application']

        # Set a value in both the application and default scopes.
        p.set('application/acme.ui.bgcolor', 'red')
        p.set('default/acme.ui.bgcolor', 'yellow')

        # Make sure when we look it up we get the one in first scope in the
        # lookup order.
        self.assertEqual('red', p.get('acme.ui.bgcolor'))

        # But we can still get at each scope individually.
        self.assertEqual('red', p.get('application/acme.ui.bgcolor'))
        self.assertEqual('yellow', p.get('default/acme.ui.bgcolor'))

        return

    def test_add_listener_in_specific_scope(self):
        """ add listener in specific scope. """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key  = key
            listener.old  = old
            listener.new  = new

            return

        # Add a listener.
        p.add_preferences_listener(listener, 'default/acme.ui')

        # Set a value and make sure the listener was called.
        p.set('default/acme.ui.bgcolor', 'blue')
        self.assertEqual(p.node('default/acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual(None, listener.old)
        self.assertEqual('blue', listener.new)

        # Set it to another value to make sure we get the 'old' value
        # correctly.
        p.set('default/acme.ui.bgcolor', 'red')
        self.assertEqual(p.node('default/acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual('blue', listener.old)
        self.assertEqual('red', listener.new)

        return

    def test_remove_listener_in_specific_scope(self):
        """ remove listener in specific scope. """

        p = self.preferences

        def listener(node, key, old, new):
            """ Listener for changes to a preferences node. """

            listener.node = node
            listener.key  = key
            listener.old  = old
            listener.new  = new

            return

        # Add a listener.
        p.add_preferences_listener(listener, 'default/acme.ui')

        # Set a value and make sure the listener was called.
        p.set('default/acme.ui.bgcolor', 'blue')
        self.assertEqual(p.node('default/acme.ui'), listener.node)
        self.assertEqual('bgcolor', listener.key)
        self.assertEqual(None, listener.old)
        self.assertEqual('blue', listener.new)

        # Remove the listener.
        p.remove_preferences_listener(listener, 'default/acme.ui')

        # Set a value and make sure the listener was *not* called.
        listener.node = None

        p.set('default/acme.ui.bgcolor', 'blue')
        self.assertEqual(None, listener.node)

        return

    def test_non_existent_scope(self):
        """ non existent scope """

        p = self.preferences

        self.assertRaises(ValueError, p.get, 'bogus/acme.ui.bgcolor')

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from apptools.permissions.adapter_base import AdapterBase
from pyface.action.api import Action


class ActionAdapter(AdapterBase):
    """This is the adapter for PyFace actions."""

    def get_enabled(self):
        """Get the enabled state of the proxied object."""

        return self.proxied.enabled

    def set_enabled(self, value):
        """Set the enabled state of the proxied object."""

        self.proxied.enabled = value

    def get_visible(self):
        """Get the visible state of the proxied object."""

        return self.proxied.visible

    def set_visible(self, value):
        """Set the visible state of the proxied object."""

        self.proxied.visible = value

    def setattr(self, name, value):
        """Reimplemented to intercept the setting of the enabled and visible
        attributes of the proxied action.
        """

        if name == 'enabled':
            self.update_enabled(value)
        elif name == 'visible':
            self.update_visible(value)
        else:
            super(ActionAdapter, self).setattr(name, value)

AdapterBase.register_adapter(ActionAdapter, Action)



import os
from contextlib import closing

import numpy as np
from numpy import testing
import tables

from ..file import H5File
from ..dict_node import H5DictNode
from ..table_node import H5TableNode
from .utils import open_h5file, temp_h5_file


H5_TEST_FILE = '_temp_test_filt.h5'


def teardown():
    try:
        os.remove(H5_TEST_FILE)
    except OSError:
        pass


def test_reopen():
    h5 = H5File(H5_TEST_FILE, mode='w')
    assert h5.is_open
    h5.close()
    assert not h5.is_open
    h5.open()
    assert h5.is_open
    h5.close()


def test_open_from_pytables_object():
    with closing(tables.File(H5_TEST_FILE, 'w')) as pyt_file:
        pyt_file.create_group('/', 'my_group')
        with open_h5file(pyt_file) as h5:
            assert '/my_group' in h5


def test_open_from_closed_pytables_object():
    with closing(tables.File(H5_TEST_FILE, 'w')) as pyt_file:
        pyt_file.create_group('/', 'my_group')
        pyt_file.close()
        with open_h5file(pyt_file) as h5:
            assert '/my_group' in h5


def test_create_array_with_H5File():
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.create_array('/array', array)
        # Test returned array
        testing.assert_allclose(h5array, array)
        # Test stored array
        testing.assert_allclose(h5['/array'], array)


def test_create_array_with_H5Group():
    array = np.arange(3)
    node_path = '/tardigrade/array'
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        group = h5.create_group('/tardigrade')
        h5array = group.create_array('array', array)
        # Test returned array
        testing.assert_allclose(h5array, array)
        # Test stored array
        testing.assert_allclose(h5[node_path], array)


def test_getitem_failure():
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_array('/array', array)
        testing.assert_raises(NameError, h5.__getitem__, '/not_there')


def test_iteritems():
    node_paths = ['/foo', '/bar', '/baz']
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        for path in node_paths:
            h5.create_array(path, array)

        # We expect to see the root node when calling iteritems...
        node_paths.append('/')
        iter_paths = []

        # 2to3 converts the iteritems blindly to items which is incorrect,
        # so we resort to this ugliness.
        items = getattr(h5, 'iteritems')()
        for path, node in items:
            iter_paths.append(path)

    assert set(node_paths) == set(iter_paths)


def test_create_plain_array_with_H5File():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.create_array('/array', np.arange(3), chunked=False)
        assert isinstance(h5array, tables.Array)
        assert not isinstance(h5array, tables.CArray)


def test_create_plain_array_with_H5Group():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.root.create_array('/array', np.arange(3), chunked=False)
        assert isinstance(h5array, tables.Array)
        assert not isinstance(h5array, tables.CArray)


def test_create_chunked_array_with_H5File():
    array = np.arange(3, dtype=np.uint8)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.create_array('/array', array, chunked=True)
        testing.assert_allclose(h5array, array)
        assert isinstance(h5array, tables.CArray)


def test_create_chunked_array_with_H5Group():
    array = np.arange(3, dtype=np.uint8)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.root.create_array('/array', array, chunked=True)
        testing.assert_allclose(h5array, array)
        assert isinstance(h5array, tables.CArray)


def test_create_extendable_array_with_H5File():
    array = np.arange(3, dtype=np.uint8)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.create_array('/array', array, extendable=True)
        testing.assert_allclose(h5array, array)
        assert isinstance(h5array, tables.EArray)


def test_create_extendable_array_with_H5Group():
    array = np.arange(3, dtype=np.uint8)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5array = h5.root.create_array('/array', array, extendable=True)
        testing.assert_allclose(h5array, array)
        assert isinstance(h5array, tables.EArray)


def test_str_and_repr():
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_array('/array', array)

        assert repr(h5) == repr(h5._h5)
        assert str(h5) == str(h5._h5)


def test_shape_and_dtype():
    array = np.ones((3, 4), dtype=np.uint8)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        for node, chunked in (('/array', False), ('/carray', True)):
            h5array = h5.create_array(node, array.shape, dtype=array.dtype,
                                      chunked=chunked)
            assert h5array.dtype == array.dtype
            assert h5array.shape == array.shape


def test_shape_only_raises():
    shape = (3, 4)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        testing.assert_raises(ValueError, h5.create_array, '/array', shape)


def test_create_duplicate_array_raises():
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=False) as h5:
        h5.create_array('/array', array)
        testing.assert_raises(ValueError, h5.create_array, '/array', array)


def test_delete_existing_array_with_H5File():
    old_array = np.arange(3)
    new_array = np.ones(5)
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_array('/array', old_array)
        # New array with the same node name should delete old array
        h5.create_array('/array', new_array)
        testing.assert_allclose(h5['/array'], new_array)


def test_delete_existing_array_with_H5Group():
    old_array = np.arange(3)
    new_array = np.ones(5)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_array('/array', old_array)
        # New array with the same node name should delete old array
        h5.root.create_array('/array', new_array, delete_existing=True)
        testing.assert_allclose(h5['/array'], new_array)


def test_delete_existing_dict_with_H5File():
    old_dict = {'a': 'Goose'}
    new_dict = {'b': 'Quail'}
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_dict('/dict', old_dict)
        # New dict with the same node name should delete old dict
        h5.create_dict('/dict', new_dict)
        assert h5['/dict'].data == new_dict


def test_delete_existing_dict_with_H5Group():
    old_dict = {'a': 'Goose'}
    new_dict = {'b': 'Quail'}
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_dict('/dict', old_dict)
        # New dict with the same node name should delete old dict
        h5.root.create_dict('/dict', new_dict, delete_existing=True)
        assert h5['/dict'].data == new_dict


def test_delete_existing_table_with_H5File():
    old_description = [('Honk', 'int'), ('Wink', 'float')]
    new_description = [('Toot', 'float'), ('Pop', 'int')]
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_table('/table', old_description)
        # New table with the same node name should delete old table
        h5.create_table('/table', new_description)
        tab = h5['/table']
        tab.append({'Pop': (1,), 'Toot': (np.pi,)})
        assert tab.ix[0][0] == np.pi


def test_delete_existing_table_with_H5Group():
    old_description = [('Honk', 'int'), ('Wink', 'float')]
    new_description = [('Toot', 'float'), ('Pop', 'int')]
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_table('/table', old_description)
        # New table with the same node name should delete old table
        h5.root.create_table('/table', new_description, delete_existing=True)
        tab = h5['/table']
        tab.append({'Pop': (1,), 'Toot': (np.pi,)})
        assert tab.ix[0][0] == np.pi


def test_delete_existing_group_with_H5File():
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_group('/group')
        grp = h5['/group']
        grp.attrs['test'] = 4

        assert grp.attrs['test'] == 4

        h5.create_group('/group')
        grp = h5['/group']
        grp.attrs['test'] = 6

        assert grp.attrs['test'] == 6


def test_delete_existing_group_with_H5Group():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_group('/group')
        grp = h5['/group']
        grp.attrs['test'] = 4

        assert grp.attrs['test'] == 4

        h5.root.create_group('/group', delete_existing=True)
        grp = h5['/group']
        grp.attrs['test'] = 6

        assert grp.attrs['test'] == 6


def test_remove_group_with_H5File():
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_group('/group')
        assert '/group' in h5
        h5.remove_group('/group')
        assert '/group' not in h5


def test_remove_group_with_H5Group():
    node_path = '/waterbear/group'
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        group = h5.create_group('/waterbear')
        group.create_group('group')
        assert node_path in h5
        group.remove_group('group')
        assert node_path not in h5


@testing.raises(ValueError)
def test_remove_group_with_remove_node():
    node_path = '/group'
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_group(node_path)
        h5.remove_node(node_path)  # Groups should be removed w/ `remove_group`


def test_remove_node_with_H5File():
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        h5.create_array('/array', np.arange(3))
        assert '/array' in h5
        h5.remove_node('/array')
        assert '/array' not in h5


def test_remove_node_with_H5Group():
    node_path = '/waterbear/array'
    with open_h5file(H5_TEST_FILE, mode='w', delete_existing=True) as h5:
        group = h5.create_group('/waterbear')
        h5.create_array(node_path, np.arange(3))
        assert node_path in h5
        group.remove_node('array')
        assert node_path not in h5


def test_read_mode_raises_on_nonexistent_file():
    cm = open_h5file('_nonexistent_.h5', mode='r')
    testing.assert_raises(IOError, cm.__enter__)
    cm = open_h5file('_nonexistent_.h5', mode='r+')
    testing.assert_raises(IOError, cm.__enter__)


def test_cleanup():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5_pytables = h5._h5  # This reference gets deleted on close
    assert not h5_pytables.isopen


def test_create_group_with_H5File():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_group('/group')
        assert '/group' in h5


def test_create_group_with_H5Group():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        group = h5['/']
        group.create_group('group')
        assert '/group' in h5


def test_split_path():
    path, node = H5File.split_path('/')
    assert path == '/'
    assert node == ''

    path, node = H5File.split_path('/node')
    assert path == '/'
    assert node == 'node'

    path, node = H5File.split_path('/group/node')
    assert path == '/group'
    assert node == 'node'


def test_join_path():
    path = H5File.join_path('/', 'a', 'b', 'c')
    assert path == '/a/b/c'

    path = H5File.join_path('a', 'b/c')
    assert path == '/a/b/c'

    path = H5File.join_path('a', '/b', '/c')
    assert path == '/a/b/c'


def test_auto_groups():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.auto_groups = True
        h5.create_array('/group/array', np.arange(3))
        assert '/group/array' in h5


def test_auto_groups_deep():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.auto_groups = True
        h5.create_array('/group1/group2/array', np.arange(3))
        assert '/group1/group2/array' in h5


def test_groups():
    array = np.arange(3)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5.create_array('/a/b/array', array)
        group_a = h5['/a']
        # Check that __contains__ works for groups
        assert 'b' in group_a
        assert 'b/array' in group_a
        group_b = group_a['b']
        # Check that __contains__ works for arrays
        assert 'array' in group_b
        testing.assert_allclose(group_b['array'], array)
        testing.assert_allclose(group_a['b/array'], array)


def test_group_attributes():
    value_1 = 'foo'
    value_2 = 'bar'
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5['/'].attrs['name'] = value_1
        assert h5['/'].attrs['name'] == value_1
        group = h5['/']

        # Make sure changes to the attribute consistent
        group._h5_group._v_attrs['name'] = value_2
        assert group.attrs['name'] == value_2


def test_group_properties():
    with open_h5file(H5_TEST_FILE, mode='w', auto_groups=True) as h5:
        h5.create_array('/group1/group2/array', np.arange(3))
        h5.create_array('/group1/array1', np.arange(3))

        assert h5['/group1'].name == 'group1'

        child_names = h5['/group1'].children_names
        assert sorted(child_names) == sorted(['group2', 'array1'])

        sub_names = h5['/group1'].subgroup_names
        assert sub_names == ['group2']

        assert h5['/group1'].root.name == '/'
        assert h5['/group1/group2'].root.name == '/'


def test_iter_groups():
    with open_h5file(H5_TEST_FILE, mode='w', auto_groups=True) as h5:
        h5.create_array('/group1/array', np.arange(3))
        h5.create_array('/group1/subgroup/deep_array', np.arange(3))
        group = h5['/group1']
        assert set(n.name for n in group.iter_groups()) == set(['subgroup'])


def test_mapping_interface_for_file():
    with open_h5file(H5_TEST_FILE, mode='w', auto_groups=True) as h5:
        array = h5.create_array('/array', np.arange(3))
        h5.create_array('/group/deep_array', np.arange(3))
        # `deep_array` isn't a direct descendent and isn't counted.
        assert len(h5) == 2
        assert '/group' in h5
        assert '/array' in h5
        testing.assert_allclose(h5['/array'], array)
        assert set(n.name for n in h5) == set(['array', 'group'])


def test_mapping_interface_for_group():
    with open_h5file(H5_TEST_FILE, mode='w', auto_groups=True) as h5:
        array = h5.create_array('/group1/array', np.arange(3))
        h5.create_array('/group1/subgroup/deep_array', np.arange(3))
        group = h5['/group1']
        # `deep_array` isn't a direct descendent and isn't counted.
        assert len(group) == 2
        assert 'subgroup' in group
        assert 'array' in group
        testing.assert_allclose(group['array'], array)
        assert set(n.name for n in group) == set(['array', 'subgroup'])


def test_group_str_and_repr():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        group = h5['/']

        assert str(group) == str(group._h5_group)
        assert repr(group) == repr(group._h5_group)


def test_attribute_translation():
    value_1 = (1, 2, 3)
    value_1_array = np.array(value_1)
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        h5['/'].attrs['name'] = value_1
        assert isinstance(h5['/'].attrs['name'], np.ndarray)
        testing.assert_allclose(h5['/'].attrs['name'], value_1_array)


def test_get_attribute():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        attrs = h5['/'].attrs
        attrs['name'] = 'hello'
        assert attrs.get('name') == attrs['name']


def test_del_attribute():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        attrs = h5['/'].attrs
        attrs['name'] = 'hello'
        del attrs['name']
        assert 'name' not in attrs


def test_get_attribute_default():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        assert h5['/'].attrs.get('missing', 'null') == 'null'


def test_attribute_update():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        attrs = h5['/'].attrs
        attrs.update({'a': 1, 'b': 2})
        assert attrs['a'] == 1
        assert attrs['b'] == 2
        attrs.update({'b': 20, 'c': 30})
        assert attrs['b'] == 20
        assert attrs['c'] == 30


def test_attribute_iteration_methods():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        attrs = h5['/'].attrs
        attrs['organ'] = 'gallbladder'
        attrs['count'] = 42
        attrs['alpha'] = 0xff

        items = list(attrs.items())

        assert all(isinstance(x, tuple) for x in items)

        # unfold the pairs
        keys, vals = list(map(list, list(zip(*items))))

        assert keys == list(attrs.keys())
        assert vals == list(attrs.values())

        # Check that __iter__ is consistent
        assert keys == list(iter(attrs))
        assert len(attrs) == len(keys)


def test_bad_node_name():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        testing.assert_raises(ValueError, h5.create_array,
                              '/attrs', np.zeros(3))


def test_bad_group_name():
    with open_h5file(H5_TEST_FILE, mode='w') as h5:
        testing.assert_raises(ValueError, h5.create_array,
                              '/attrs/array', np.zeros(3))


def test_create_dict_with_H5File():
    data = {'a': 1}
    with temp_h5_file() as h5:
        h5.create_dict('/dict', data)
        assert isinstance(h5['/dict'], H5DictNode)
        assert h5['/dict']['a'] == 1


def test_create_dict_with_H5Group():
    node_path = '/bananas/dict'
    data = {'a': 1}
    with temp_h5_file() as h5:
        group = h5.create_group('/bananas')
        group.create_dict('dict', data)
        assert isinstance(h5[node_path], H5DictNode)
        assert h5[node_path]['a'] == 1


def test_create_table_with_H5File():
    description = [('foo', 'int'), ('bar', 'float')]
    with temp_h5_file() as h5:
        h5.create_table('/table', description)
        tab = h5['/table']
        assert isinstance(tab, H5TableNode)
        tab.append({'foo': (1,), 'bar': (np.pi,)})
        assert tab.ix[0][0] == 1
        assert tab.ix[0][1] == np.pi

        h5.remove_node('/table')
        assert '/table' not in h5


def test_create_table_with_H5Group():
    node_path = '/rhinocerous/table'
    description = [('foo', 'int'), ('bar', 'float')]
    with temp_h5_file() as h5:
        group = h5.create_group('/rhinocerous')
        group.create_table('table', description)
        tab = h5[node_path]
        assert isinstance(tab, H5TableNode)
        tab.append({'foo': (1,), 'bar': (np.pi,)})
        assert tab.ix[0][0] == 1
        assert tab.ix[0][1] == np.pi

        group.remove_node('table')
        assert node_path not in h5


if __name__ == '__main__':
    testing.run_module_suite()



"""Public API for the scripting package.
"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2008, Prabhu Ramachandran and Enthought, Inc.
# License: BSD Style.

from .recorder import Recorder, RecorderError
from .recordable import recordable
from .package_globals import get_recorder, set_recorder
from .recorder_with_ui import RecorderWithUI
from .util import start_recording, stop_recording



""" Tests for Python-esque '.ini' files. """


# Standard library imports.
import os, tempfile, unittest
from os.path import join

# Major package imports.
from pkg_resources import resource_filename

# Enthought library imports.
from .py_config_file import PyConfigFile


# This module's package.
PKG = 'apptools.preferences.tests'


class PyConfigFileTestCase(unittest.TestCase):
    """ Tests for Python-esque '.ini' files. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """

        # The filenames of the example preferences files.
        self.example = resource_filename(PKG, 'py_config_example.ini')
        self.example_2 = resource_filename(PKG, 'py_config_example_2.ini')

        return

    def tearDown(self):
        """ Called immediately after each test method has been called. """

        return

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_load_from_filename(self):
        """ load from filename """

        config = PyConfigFile(self.example)

        self.assertEqual('blue', config['acme.ui']['bgcolor'])
        self.assertEqual(50, config['acme.ui']['width'])
        self.assertEqual(1.0, config['acme.ui']['ratio'])
        self.assertEqual(True, config['acme.ui']['visible'])
        self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
        self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
        self.assertEqual((1, 'a', 6, 4), config['acme.ui']['baz'])
        self.assertEqual('red', config['acme.ui.other']['fred'])
        self.assertEqual(100, config['acme.ui.other']['wilma'])
        self.assertEqual(90, config['tds.foogle']['joe'])
        self.assertEqual("meerkat", config['simples']['animal'])

        return

    def test_load_from_file(self):
        """ load from file """

        config = PyConfigFile(open(self.example))

        self.assertEqual('blue', config['acme.ui']['bgcolor'])
        self.assertEqual(50, config['acme.ui']['width'])
        self.assertEqual(1.0, config['acme.ui']['ratio'])
        self.assertEqual(True, config['acme.ui']['visible'])
        self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
        self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
        self.assertEqual((1, 'a', 6, 4), config['acme.ui']['baz'])
        self.assertEqual('red', config['acme.ui.other']['fred'])
        self.assertEqual(100, config['acme.ui.other']['wilma'])
        self.assertEqual(90, config['tds.foogle']['joe'])
        self.assertEqual("meerkat", config['simples']['animal'])

        return

    def test_save(self):
        """ save """

        config = PyConfigFile(open(self.example))

        self.assertEqual('blue', config['acme.ui']['bgcolor'])
        self.assertEqual(50, config['acme.ui']['width'])
        self.assertEqual(1.0, config['acme.ui']['ratio'])
        self.assertEqual(True, config['acme.ui']['visible'])
        self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
        self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
        self.assertEqual('red', config['acme.ui.other']['fred'])
        self.assertEqual(100, config['acme.ui.other']['wilma'])
        self.assertEqual(90, config['tds.foogle']['joe'])
        self.assertEqual("meerkat", config['simples']['animal'])

        # Save the config to another file.
        tmpdir = tempfile.mkdtemp()
        tmp    = join(tmpdir, 'tmp.ini')

        config.save(tmp)
        try:
            self.assertTrue(os.path.exists(tmp))

            # Make sure we can read the file back in and that we get the same
            # values!
            config = PyConfigFile(open(tmp))

            self.assertEqual('blue', config['acme.ui']['bgcolor'])
            self.assertEqual(50, config['acme.ui']['width'])
            self.assertEqual(1.0, config['acme.ui']['ratio'])
            self.assertEqual(True, config['acme.ui']['visible'])
            self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
            self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
            self.assertEqual((1, 'a', 6, 4), config['acme.ui']['baz'])
            self.assertEqual('red', config['acme.ui.other']['fred'])
            self.assertEqual(100, config['acme.ui.other']['wilma'])
            self.assertEqual(90, config['tds.foogle']['joe'])
            self.assertEqual("meerkat", config['simples']['animal'])

        finally:
            # Clean up!
            os.remove(tmp)
            os.removedirs(tmpdir)

        return

    def test_load_multiple_files(self):
        """ load multiple files """

        config = PyConfigFile(self.example)

        self.assertEqual('blue', config['acme.ui']['bgcolor'])
        self.assertEqual(50, config['acme.ui']['width'])
        self.assertEqual(1.0, config['acme.ui']['ratio'])
        self.assertEqual(True, config['acme.ui']['visible'])
        self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
        self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
        self.assertEqual((1, 'a', 6, 4), config['acme.ui']['baz'])
        self.assertEqual('red', config['acme.ui.other']['fred'])
        self.assertEqual(100, config['acme.ui.other']['wilma'])
        self.assertEqual(90, config['tds.foogle']['joe'])
        self.assertEqual("meerkat", config['simples']['animal'])

        # Load another file.
        config.load(self.example_2)

        # Make sure we still have the unchanged values...
        self.assertEqual('red', config['acme.ui']['bgcolor'])
        self.assertEqual(50, config['acme.ui']['width'])
        self.assertEqual(1.0, config['acme.ui']['ratio'])
        self.assertEqual(True, config['acme.ui']['visible'])
        self.assertEqual({'a' : 1, 'b' : 2}, config['acme.ui']['foo'])
        self.assertEqual([1, 2, 3, 4], config['acme.ui']['bar'])
        self.assertEqual((1, 'a', 6, 4), config['acme.ui']['baz'])
        self.assertEqual('red', config['acme.ui.other']['fred'])
        self.assertEqual(100, config['acme.ui.other']['wilma'])
        self.assertEqual(90, config['tds.foogle']['joe'])
        self.assertEqual("meerkat", config['simples']['animal'])

        # ... and the values that were overwritten...
        self.assertEqual('red', config['acme.ui']['bgcolor'])

        # ... and that we have the new ones.
        self.assertEqual(42, config['acme.ui']['bazzle'])

        # ... and that the new ones can refer to the old ones!
        self.assertEqual(180, config['acme.ui']['blimey'])

        return


# Entry point for stand-alone testing.
if __name__ == '__main__':
    unittest.main()

#### EOF ######################################################################



#-------------------------------------------------------------------------------
#
#  Defines the TemplateDataName class used for binding a data source to a data
#  context.
#
#  Written by: David C. Morrill
#
#  Date: 07/27/2007
#
#  (c) Copyright 2007 by Enthought, Inc.
#
#-------------------------------------------------------------------------------

""" Defines the TemplateDataName class used for binding a data source to a data
    context.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Instance, Bool, Property, Undefined, \
           on_trait_change, cached_property

from .itemplate_data_name_item \
    import ITemplateDataNameItem

from .itemplate_data_context \
    import ITemplateDataContext

from .template_impl \
    import Template

from .template_traits \
    import TList, TStr, TBool

#-------------------------------------------------------------------------------
#  'TemplateDataName' interface:
#-------------------------------------------------------------------------------

class TemplateDataName ( Template ):
    """ Defines the TemplateDataName class used for binding a data source to a
        data context.
    """

    #-- Public Template Traits -------------------------------------------------

    # The list of ITemplateDataNameItem's used to match the data context:
    items = TList( ITemplateDataNameItem )

    # A description of the template data name (for use in a user interface):
    description = TStr

    # Is this binding optional?
    optional = TBool( False )

    #-- Public Non-Template Traits ---------------------------------------------

    # The data context the template name is matching against currently:
    context = Instance( ITemplateDataContext )

    # Is the binding resolved?
    resolved = Property( Bool, depends_on = 'items.output_data_context' )

    # The actual name of the context data that the data source is bound to:
    context_data_name = Property

    # The context data the data source is bound to:
    context_data = Property

    #-- Property Implementations -----------------------------------------------

    @cached_property
    def _get_resolved ( self ):
        if len( self.items ) == 0:
            return False

        context = self.items[-1].output_data_context
        return ((context is not None) and
                (len( context.data_context_values ) == 1) and
                (len( context.data_contexts ) == 0))

    def _get_context_data_name ( self ):
        if self.resolved:
            context = self.items[-1].output_data_context
            path    = context.data_context_path
            if path != '':
                path += '.'
            name = context.data_context_name
            if name != '':
                name += '.'

            return '%s%s%s' % ( path, name, context.data_context_values[0] )

        return ''

    def _get_context_data ( self ):
        if self.resolved:
            context = self.items[-1].output_data_context
            return context.get_data_context_value(
                       context.data_context_values[0] )

        return Undefined

    #-- Trait Event Handlers ---------------------------------------------------

    def _context_changed ( self, context ):
        """ Resets the name whenever the context is changed.
        """
        self._reset()

    @on_trait_change( ' items' )
    def _on_items_changed ( self ):
        """ Resets the name whenever any of the name items are changed.
        """
        self._reset()

    @on_trait_change( 'items:output_data_context' )
    def _on_output_data_context_changed ( self, item, name, old, new ):
        i = self.items.index( item )
        if i < (len( self.items ) - 1):
            self.items[ i + 1 ].input_data_context = new

    #-- Private Methods --------------------------------------------------------

    def _reset ( self ):
        """ Resets the name whenever any significant change occurs.
        """
        items = self.items
        n     = len( items )
        if n > 0:
            items[0].input_data_context = self.context
            for i in range( 0, n - 1 ):
                items[ i + 1 ].input_data_context = \
                    items[ i ].output_data_context




from contextlib import closing
import json

from numpy import ndarray

from tables import Group as PyTablesGroup
from tables.nodes import filenode


#: The key name which identifies array objects in the JSON dict.
ARRAY_PROXY_KEY = '__array__'
NODE_KEY = 'node_name'


class H5DictNode(object):
    """ Dictionary-like node interface.

    Data for the dict is stored as a JSON file in a PyTables FileNode. This
    allows easy storage of Python objects, such as dictionaries and lists of
    different data types.

    Note that this is implemented using a group-node assuming that arrays are
    valid inputs and will be stored as H5 array nodes.

    Parameters
    ----------
    h5_group : H5Group instance
        Group node which will be used as a dictionary store.
    auto_flush : bool
        If True, write data to disk whenever the dict data is altered.
        Otherwise, call `flush()` explicitly to write data to disk.
    """

    #: Name of filenode where dict data is stored.
    _pyobject_data_node = '_pyobject_data'

    def __init__(self, h5_group, auto_flush=True):
        assert self.is_dict_node(h5_group)

        h5_group = self._get_pyt_group(h5_group)
        self._h5_group = h5_group
        self.auto_flush = auto_flush

        # Load dict data from the file node.
        dict_node = getattr(h5_group, self._pyobject_data_node)
        with closing(filenode.open_node(dict_node)) as f:
            self._pyobject_data = json.loads(
                f.read().decode('ascii'), object_hook=self._object_hook
            )

    #--------------------------------------------------------------------------
    #  Dictionary interface
    #--------------------------------------------------------------------------

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value
        if self.auto_flush:
            self.flush()

    def __delitem__(self, key):
        del self.data[key]
        if self.auto_flush:
            self.flush()

    def __contains__(self, key):
        return key in self.data

    def keys(self):
        return list(self.data.keys())

    #--------------------------------------------------------------------------
    #  Public interface
    #--------------------------------------------------------------------------

    @property
    def data(self):
        return self._pyobject_data

    @data.setter
    def data(self, new_data_dict):
        self._pyobject_data = new_data_dict
        if self.auto_flush:
            self.flush()

    def flush(self):
        """ Write buffered data to disk. """
        self._remove_pyobject_node()
        self._write_pyobject_node()

    @classmethod
    def add_to_h5file(cls, h5, node_path, data=None, **kwargs):
        """ Add dict node to an H5 file at the specified path.

        Parameters
        ----------
        h5 : H5File
            The H5 file where the dictionary data will be stored.
        node_path : str
            Path to node where data is stored (e.g. '/path/to/my_dict')
        data : dict
            Data for initialization, if desired.
        """
        h5.create_group(node_path)
        group = h5[node_path]

        cls._create_pyobject_node(h5._h5, node_path, data=data)
        return cls(group, **kwargs)

    @classmethod
    def is_dict_node(cls, pytables_node):
        """ Return True if PyTables node looks like an H5DictNode.

        NOTE: That this returns False if the node is an `H5DictNode` instance,
        since the input node should be a normal PyTables Group node.
        """
        # Import here to prevent circular imports
        from .file import H5Group

        if isinstance(pytables_node, H5Group):
            pytables_node = cls._get_pyt_group(pytables_node)

        if not isinstance(pytables_node, PyTablesGroup):
            return False

        return cls._pyobject_data_node in pytables_node._v_children

    #--------------------------------------------------------------------------
    #  Private interface
    #--------------------------------------------------------------------------

    def _f_remove(self):
        """ This is called by H5File whenever a node is removed.

        All nodes in `_h5_group` will be removed.
        """
        for name in list(self._h5_group._v_children.keys()):
            if name != self._pyobject_data_node:
                self._h5_group.__getattr__(name)._f_remove()
        # Remove the dict node
        self._remove_pyobject_node()
        # Remove the group node
        self._h5_group._f_remove()

    def _object_hook(self, dct):
        """ This gets passed object dictionaries by `json.load(s)` and if it
        finds `ARRAY_PROXY_KEY` in the object description it returns the
        proxied array object.
        """
        if ARRAY_PROXY_KEY in dct:
            node_name = dct[NODE_KEY]
            return getattr(self._h5_group, node_name)[:]
        return dct

    def _remove_pyobject_node(self):
        node = getattr(self._h5_group, self._pyobject_data_node)
        node._f_remove()

    def _write_pyobject_node(self):
        pyt_file = self._h5_group._v_file
        node_path = self._h5_group._v_pathname
        self._create_pyobject_node(pyt_file, node_path, self.data)

    @classmethod
    def _create_pyobject_node(cls, pyt_file, node_path, data=None):
        if data is None:
            data = {}

        # Stash the array values in their own h5 nodes and return a dictionary
        # which is appropriate for JSON serialization.
        out_data = cls._handle_array_values(pyt_file, node_path, data)

        kwargs = dict(where=node_path, name=cls._pyobject_data_node)
        with closing(filenode.new_node(pyt_file, **kwargs)) as f:
            f.write(json.dumps(out_data).encode('ascii'))

    @classmethod
    def _get_pyt_group(self, group):
        if hasattr(group, '_h5_group'):
            group = group._h5_group
        return group

    @classmethod
    def _array_proxy(cls, pyt_file, group, key, array):
        """ Stores an array as a normal H5 node and returns the proxy object
        which will be serialized to JSON.

        `ARRAY_PROXY_KEY` marks the object dictionary as an array proxy so that
        `_object_hook` can recognize it. `NODE_KEY` stores the node name of the
        array so that `_object_hook` can load the array data when the dict node
        is deserialized.

        """
        if key in group:
            pyt_file.remove_node(group, key)
        pyt_file.create_array(group, key, array)
        return {ARRAY_PROXY_KEY: True, NODE_KEY: key}

    @classmethod
    def _handle_array_values(cls, pyt_file, group_path, data):
        group = pyt_file.get_node(group_path)

        # Convert numpy array values to H5 array nodes.
        out_data = {}
        for key in list(data.keys()):
            value = data[key]
            if isinstance(value, ndarray):
                out_data[key] = cls._array_proxy(pyt_file, group, key, value)
            else:
                out_data[key] = value

        # Remove stored arrays which are no longer in the data dictionary.
        pyt_children = group._v_children
        for key in list(pyt_children.keys()):
            if key not in data:
                pyt_file.remove_node(group, key)

        return out_data



#-----------------------------------------------------------------------------
#
#  Copyright (c) 2006 by Enthought, Inc.
#  All rights reserved.
#
#  Author: Dave Peterson <dpeterson@enthought.com>
#
#-----------------------------------------------------------------------------

""" A pickle framework that supports unpickling of refactored versions of
    old classes. Part of the AppTools project of the Enthought Tool Suite.

    Beyond refactoring support, there are additional useful capabilities of
    this package:

    - HasTraits objects are fully pickled when sweet_pickle.dump() or
      sweet_pickle.dumps() are called.  This means every trait's value
      is pickled explicitly.  This allows improved self-consistency as
      values within unpickled objects correspond to the default trait
      values of the instance that was pickled rather than the default
      values of the current class definition.
    - Multiple updaters can be registered to update state for any version
      of a class.  This is extremely useful when an instance being
      pickled and unpickled is the result of an addition of one or more
      Trait Categories to a class.  In that case, the Category contributor
      can also register a state function to update the Category's traits
      through versioning.


    We have duplicated the API of the Python 'pickle' module within this
    package. Thus, users can simply import sweet_pickle in places where
    they previously used pickle or cPickle.  For example::

        import cPickle          --->    import apptools.sweet_pickle as pickle
        s = pickle.dumps(obj)           s = pickle.dumps(obj)
        pickle.loads(s)                 pickle.loads(s)

    Pickles generated by this package *can be* unpickled by standard pickle
    or cPickle, though you lose the benefit of the refactoring support during
    unpickling.

    As a result of the above, this package is a drop-in replacement for the
    Python 'pickle' module or 'cPickle' module and should be safe to use even
    if you never version or refactor your own classes.  In fact, we STRONGLY
    RECOMMEND that you use this framework for all of your pickle needs because
    you never know when one of the classes you encounter during unpickling has
    been versioned or otherwise refactored by someone else.

    See module 'pickle' for more basic information about pickling.


    The most common way to benefit from the versioning capabilities of this
    framework is to register class mappings and state modification functions
    with the global updater registry (more detail is below.)  However, you may
    also choose to explicitly instantiate an Unpickler and provide it with your
    own explicit definition of class mappings and state modification functions.
    We do not provide any help on the latter at this time.

    You can register class mappings and state modification functions with the
    global updater registry using the methods provided on the Updater instance
    that is the global registry.  You can get this instance by calling the
    'get_global_registry' function exposed through this package's namespace.

    This framework has been designed so that you can register your class
    mappings and state modification functions during the import of the module
    or package that contained the original class.  However, you may also
    register your mappings and functions at any point such that the they are
    known to the framework prior to, or become known during the attempt to,
    unpickle the class they modify.

    The framework will call a __setstate__ method on the final target class
    of any unpickled instance.  It will not call __setstate__ methods on any
    beginning or intermediate classes within a chain of class mappings.


    A class mapping is used to redirect the unpickling of one class to return
    an instantiation of another class.  The classes can be in different
    modules, and the modules in different packages.  Mappings can be chained.
    For example, given the mappings::

        foo.bar.Bar --> foo.baz.Baz
        foo.baz.Baz --> foo.Foo

    An attempt to unpickle a foo.bar.Bar would actually generate a foo.Foo
    instance.

    A state modification function is called during the execution of the
    __setstate__ method during unpickling of an object of the type and version
    for which it was registered for.  The function must accept a single
    argument which is a state dictionary and must then return the modified
    state dictionary.  Additionally, the function should change the version
    variable within the state to represent the version the new state
    represents.  (The framework will assume an increment of one if this is
    not done.)  The framework ensures that state modification functions
    are chained appropriately to convert through multiple versions and/or class
    mappings.

    Note that refactorings that cause classes to be completed removed from
    the source code can be supported, without breaking unpickling of object
    hierarchies that include an instace of that class, by adding a mapping
    to the Placeholder class in the placeholder module.
"""

##############################################################################
# Implement the Python pickle package API
##############################################################################

# Expose our custom pickler as the standard Unpickler
from .versioned_unpickler import VersionedUnpickler as Unpickler

# Use our custom unpickler to load from files
def load(file, max_pass=-1):
    return Unpickler(file).load(max_pass)

# Use our custom unpickler to load from strings
def loads(str, max_pass=-1):
    from io import BytesIO
    file = BytesIO(str)

    return Unpickler(file).load(max_pass)

# We don't customize the Python pickler, though we do use the cPickle module
# for improved performance.
from pickle import Pickler

# Implement the dump and dumps methods so that all traits in a HasTraits object
# get included in the pickle.
def dump(obj, file, protocol=2):
    _flush_traits(obj)
    from pickle import dump as d
    return d(obj, file, protocol)

def dumps(obj, protocol=2):
    _flush_traits(obj)
    from pickle import dumps as ds
    return ds(obj, protocol)

# We don't customize exceptions so just map to the Python pickle package
from pickle import PickleError, PicklingError, UnpicklingError


##############################################################################
# Allow retrieval of the global registry
##############################################################################

from .global_registry import get_global_registry


##############################################################################
# Expose our Updater class so users can explicitly create their own.
##############################################################################

from .updater import Updater


##############################################################################
# Method to ensure that all traits on a has traits object are included in a
# pickle.
##############################################################################

def _flush_traits(obj):
    if hasattr(obj, 'trait_names'):
        for name, value in list(obj.traits().items()):
            if value.type == 'trait':
                try:
                    getattr(obj, name)
                except AttributeError:
                    pass


### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------



from .command_action import CommandAction
from .redo_action import RedoAction
from .undo_action import UndoAction



"""A version registry that manages handlers for different state
versions.
"""
# Author: Prabhu Ramachandran <prabhu_r@users.sf.net>
# Copyright (c) 2005, Enthought, Inc.
# License: BSD Style.

# Standard library imports.
import sys
import inspect
import logging


logger = logging.getLogger(__name__)


######################################################################
# Utility functions.
######################################################################
def get_version(obj):
    """Walks the class hierarchy and obtains the versions of the
    various classes and returns a list of tuples of the form
    ((class_name, module), version) in reverse order of the MRO.
    """
    res = []
    for cls in inspect.getmro(obj.__class__):
        class_name, module = cls.__name__, cls.__module__
        if module in ['__builtin__']:
            # No point in versioning builtins.
            continue
        try:
            version = cls.__version__
        except AttributeError:
            version = -1
        res.append( ( (class_name, module), version) )
    res.reverse()
    return res


######################################################################
# `HandlerRegistry` class.
######################################################################
class HandlerRegistry:
    """A simple version conversion handler registry.  Classes register
    handlers in order to convert the state version to the latest
    version.  When an object's state is about to be set, the `update`
    method of the registy is called.  This in turn calls any handlers
    registered for the class/module and this handler is then called
    with the state and the version of the state.  The state is
    modified in-place by the handlers.
    """

    def __init__(self):
        # The version conversion handlers.
        # Key: (class_name, module), value: handler
        self.handlers = {}

    def register(self, class_name, module, handler):
        """Register `handler` that handles versioning for class having
        class name (`class_name`) and module name (`module`).  The
        handler function will be passed the state and its version to fix.
        """
        key = (class_name, module)
        if key in self.handlers:
            msg = 'Overwriting version handler for (%s, %s)'%(key[0], key[1])
            logger.warn(msg)
        self.handlers[(class_name, module)] = handler

    def unregister(self, class_name, module):
        """Unregisters any handlers for a class and module.
        """
        self.handlers.pop((class_name, module))

    def update(self, state):
        """Updates the given state using the handlers.  Note that the
        state is modified in-place.
        """
        if (not self.handlers) or  (not hasattr(state, '__metadata__')):
            return
        versions = state.__metadata__['version']
        for ver in versions:
            key = ver[0]
            try:
                self.handlers[key](state, ver[1])
            except KeyError:
                pass


def _create_registry():
    """Creates a reload safe, singleton registry.
    """
    registry = None
    for key in list(sys.modules.keys()):
        if 'version_registry' in key:
            mod = sys.modules[key]
            if hasattr(mod, 'registry'):
                registry = mod.registry
                break
    if not registry:
        registry = HandlerRegistry()
    return registry


# The singleton registry.
registry = _create_registry()




#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A monitor that detects changes to a naming context. """


# Enthought library imports.
from pyface.tree.api import NodeMonitor


class ContextMonitor(NodeMonitor):
    """ A monitor that detects changes to a naming context. """

    ###########################################################################
    # 'NodeMonitor' interface.
    ###########################################################################

    def start(self):
        """ Start listening to changes to the object. """

        context = self.node.obj

        context.on_trait_change(self._on_object_added, 'object_added')
        context.on_trait_change(self._on_object_changed, 'object_changed')
        context.on_trait_change(self._on_object_removed, 'object_removed')
        context.on_trait_change(self._on_object_renamed, 'object_renamed')
        context.on_trait_change(self._on_context_changed, 'context_changed')

        return

    def stop(self):
        """ Stop listening to changes to the object. """

        context = self.node.obj

        context.on_trait_change(
            self._on_object_added, 'object_added', remove=True
        )

        context.on_trait_change(
            self._on_object_changed, 'object_changed', remove=True
        )

        context.on_trait_change(
            self._on_object_removed, 'object_removed', remove=True
        )

        context.on_trait_change(
            self._on_object_renamed, 'object_renamed', remove=True
        )

        context.on_trait_change(
            self._on_context_changed, 'context_changed', remove=True
        )

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait event handlers #################################################

    def _on_object_added(self, event):
        """ Called when an object has been added to the context. """

        self.fire_nodes_inserted([event.new_binding])

        return

    def _on_object_changed(self, event):
        """ Called when an object in the context has been changed. """

        # fixme: Can we get enough information to fire a 'nodes_replaced'
        # event?  Something a little more granular than this!
        self.fire_structure_changed()

        return

    def _on_object_removed(self, event):
        """ Called when an object has been removed from the context. """

        self.fire_nodes_removed([event.old_binding])

        return

    def _on_object_renamed(self, event):
        """ Called when an object has been renamed in the context. """

        self.fire_nodes_replaced([event.old_binding], [event.new_binding])

        return

    def _on_context_changed(self, event):
        """ Called when a context has changed dramatically. """
        self.fire_structure_changed()

        return

##### EOF #####################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
#------------------------------------------------------------------------------



# Enthought library imports.
from pyface.action.api import Action
from traits.api import Instance

# Local library imports
from ..i_undo_manager import IUndoManager


class AbstractCommandStackAction(Action):
    """ The abstract base class for all actions that operate on a command
    stack.
    """

    #### 'AbstractCommandStackAction' interface ###############################

    # The undo manager.
    undo_manager = Instance(IUndoManager)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Initialise the instance. """

        super(AbstractCommandStackAction, self).__init__(**traits)

        self.undo_manager.on_trait_event(self._on_stack_updated, 'stack_updated')

        # Update the action to initialise it.
        self._update_action()

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def destroy(self):
        """ Called when the action is no longer required.

        By default this method does nothing, but this would be a great place to
        unhook trait listeners etc.

        """

        self.undo_manager.on_trait_event(self._on_stack_updated,
                                         'stack_updated', remove=True)

    ###########################################################################
    # Protected interface.
    ###########################################################################

    def _update_action(self):
        """ Update the state of the action. """

        raise NotImplementedError

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _on_stack_updated(self, stack):
        """ Handle changes to the state of a command stack. """

        # Ignore unless it is the active stack.
        if stack is self.undo_manager.active_stack:
            self._update_action()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" A Python File System context. """


# Standard library imports.
import pickle, glob, logging, os
from os.path import join, splitext

# Enthought library imports.
from apptools.io.api import File
from traits.api import Any, Dict, Instance, Property, Str

# Local imports.
from .address import Address
from .binding import Binding
from .context import Context
from .dir_context import DirContext
from .exception import NameNotFoundError, NotContextError
from .naming_event import NamingEvent
from .naming_manager import naming_manager
from .object_serializer import ObjectSerializer
from .pyfs_context_factory import PyFSContextFactory
from .pyfs_object_factory import PyFSObjectFactory
from .pyfs_state_factory import PyFSStateFactory
from .reference import Reference
from .referenceable import Referenceable


# Setup a logger for this module.
logger = logging.getLogger(__name__)


# The name of the 'special' file in which we store object attributes.
ATTRIBUTES_FILE = '__attributes__'

# Constants for environment property keys.
FILTERS = "apptools.naming.pyfs.filters"
OBJECT_SERIALIZERS = "apptools.naming.pyfs.object.serializers"


# The default environment.
ENVIRONMENT = {
    #### 'Context' properties #################################################

    # Object factories.
    Context.OBJECT_FACTORIES : [PyFSObjectFactory(), PyFSContextFactory()],

    # State factories.
    Context.STATE_FACTORIES  : [PyFSStateFactory()],

    #### 'PyFSContext' properties #############################################

    # Object serializers.
    OBJECT_SERIALIZERS : [ObjectSerializer()],

    # List of filename patterns to ignore.  These patterns are passed to
    # 'glob.glob', so things like '*.pyc' will do what you expect.
    #
    # fixme: We should have a generalized filter mechanism here, and '.svn'
    # should be moved elsewhere!
    FILTERS : [ATTRIBUTES_FILE, '.svn']
}


class PyFSContext(DirContext, Referenceable):
    """ A Python File System context.

    This context represents a directory on a local file system.

    """

    # The name of the 'special' file in which we store object attributes.
    ATTRIBUTES_FILE = ATTRIBUTES_FILE

    # Environment property keys.
    FILTERS = FILTERS
    OBJECT_SERIALIZERS = OBJECT_SERIALIZERS

    #### 'Context' interface ##################################################

    # The naming environment in effect for this context.
    environment = Dict(ENVIRONMENT)

    # The name of the context within its own namespace.
    namespace_name = Property(Str)

    #### 'PyFSContext' interface ##############################################

    # The name of the context (the last component of the path).
    name = Str

    # The path name of the directory on the local file system.
    path = Str

    #### 'Referenceable' interface ############################################

    # The object's reference suitable for binding in a naming context.
    reference = Property(Instance(Reference))

    #### Private interface ####################################################

    # A mapping from bound name to the name of the corresponding file or
    # directory on the file system.
    _name_to_filename_map = Dict#(Str, Str)

    # The attributes of every object in the context.  The attributes for the
    # context itself have the empty string as the key.
    #
    # {str name : dict attributes}
    #
    # fixme: Don't use 'Dict' here as it causes problems when pickling because
    # trait dicts have a reference back to the parent object (hence we end up
    # pickling all kinds of things that we don't need or want to!).
    _attributes = Any

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Creates a new context. """

        # Base class constructor.
        super(PyFSContext, self).__init__(**traits)

        # We cache each object as it is looked up so that all accesses to a
        # serialized Python object return a reference to exactly the same one.
        self._cache = {}

        return

    ###########################################################################
    # 'PyFSContext' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_namespace_name(self):
        """ Returns the name of the context within its own namespace. """

        # fixme: clean this up with an initial context API!
        if 'root' in self.environment:
            root = self.environment['root']

            namespace_name = self.path[len(root) + 1:]

        else:
            namespace_name = self.path

        # fixme: This is a bit dodgy 'cos we actually return a name that can
        # be looked up, and not the file system name...
        namespace_name = '/'.join(namespace_name.split(os.path.sep))

        return namespace_name

    #### methods ##############################################################

    def refresh(self):
        """ Refresh the context to reflect changes in the file system. """

        # fixme: This needs more work 'cos if we refresh a context then we
        # will load new copies of serialized Python objects!

        # This causes the initializer to run again the next time the trait is
        # accessed.
        self.reset_traits(['_name_to_filename_map'])

        # Clear out the cache.
        self._cache = {}

        # fixme: This is a bit hacky since the context in the binding may
        # not be None!
        self.context_changed = NamingEvent(
            new_binding=Binding(name=self.name, obj=self, context=None)
        )

        return

    ###########################################################################
    # 'Referenceable' interface.
    ###########################################################################

    #### Properties ###########################################################

    def _get_reference(self):
        """ Returns a reference to this object suitable for binding. """

        abspath = os.path.abspath(self.path)

        reference = Reference(
            class_name = self.__class__.__name__,
            addresses  = [Address(type='pyfs_context', content=abspath)]
        )

        return reference

    ###########################################################################
    # Protected 'Context' interface.
    ###########################################################################

    def _is_bound(self, name):
        """ Is a name bound in this context? """

        return name in self._name_to_filename_map

    def _lookup(self, name):
        """ Looks up a name in this context. """

        if name in self._cache:
            obj = self._cache[name]

        else:
            # Get the full path to the file.
            path = join(self.path, self._name_to_filename_map[name])

            # If the file contains a serialized Python object then load it.
            for serializer in self._get_object_serializers():
                if serializer.can_load(path):
                    try:
                        state = serializer.load(path)

                    # If the load fails then we create a generic file resource
                    # (the idea being that it might be useful to have access to
                    # the file to see what went wrong).
                    except:
                        state = File(path)
                        logger.exception('Error loading resource at %s' % path)

                    break

            # Otherwise, it must just be a file or folder.
            else:
                # Directories are contexts.
                if os.path.isdir(path):
                    state = self._context_factory(name, path)

                # Files are just files!
                elif os.path.isfile(path):
                    state = File(path)

                else:
                    raise ValueError('unrecognized file for %s' % name)

            # Get the actual object from the naming manager.
            obj = naming_manager.get_object_instance(state, name, self)

            # Update the cache.
            self._cache[name] = obj

        return obj

    def _bind(self, name, obj):
        """ Binds a name to an object in this context. """

        # Get the actual state to bind from the naming manager.
        state = naming_manager.get_state_to_bind(obj, name, self)

        # If the object is actually an abstract file then we don't have to
        # do anything.
        if isinstance(state, File):
            if not state.exists:
                state.create_file()

            filename = name

        # Otherwise we are binding an arbitrary Python object, so find a
        # serializer for it.
        else:
            for serializer in self._get_object_serializers():
                if serializer.can_save(obj):
                    path = serializer.save(join(self.path, name), obj)
                    filename = os.path.basename(path)
                    break

            else:
                raise ValueError('cannot serialize object %s' % name)

        # Update the name to filename map.
        self._name_to_filename_map[name] = filename

        # Update the cache.
        self._cache[name] = obj

        return state

    def _rebind(self, name, obj):
        """ Rebinds a name to an object in this context. """

        # We unbind first to make sure that the old file gets removed (this
        # is handy if the object that we are rebinding has a different
        # serializer than the current one).
        #self._unbind(name)

        self._bind(name, obj)

        return

    def _unbind(self, name):
        """ Unbinds a name from this context. """

        # Get the full path to the file.
        path = join(self.path, self._name_to_filename_map[name])

        # Remove it!
        f = File(path)
        f.delete()

        # Update the name to filename map.
        del self._name_to_filename_map[name]

        # Update the cache.
        if name in self._cache:
            del self._cache[name]

        # Remove any attributes.
        if name in self._attributes:
            del self._attributes[name]
            self._save_attributes()

        return

    def _rename(self, old_name, new_name):
        """ Renames an object in this context. """

        # Get the old filename.
        old_filename = self._name_to_filename_map[old_name]
        old_file = File(join(self.path, old_filename))

        # Lookup the object bound to the old name.  This has the side effect
        # of adding the object to the cache under the name 'old_name'.
        obj = self._lookup(old_name)

        # We are renaming a LOCAL context (ie. a folder)...
        if old_file.is_folder:
            # Create the new filename.
            new_filename = new_name
            new_file = File(join(self.path, new_filename))

            # Move the folder.
            old_file.move(new_file)

            # Update the 'Context' object.
            obj.path = new_file.path

            # Update the cache.
            self._cache[new_name] = obj
            del self._cache[old_name]

            # Refreshing the context makes sure that all of its contents
            # reflect the new name (i.e., sub-folders and files have the
            # correct path).
            #
            # fixme: This currently results in new copies of serialized
            # Python objects!  We need to be a bit more judicious in the
            # refresh.
            obj.refresh()

        # We are renaming a file...
        elif isinstance(obj, File):
            # Create the new filename.
            new_filename = new_name
            new_file = File(join(self.path, new_filename))

            # Move the file.
            old_file.move(new_file)

            # Update the 'File' object.
            obj.path = new_file.path

            # Update the cache.
            self._cache[new_name] = obj
            del self._cache[old_name]

        # We are renaming a serialized Python object...
        else:
            # Create the new filename.
            new_filename = new_name + old_file.ext
            new_file = File(join(self.path, new_filename))

            old_file.delete()

            # Update the cache.
            if old_name in self._cache:
                self._cache[new_name] = self._cache[old_name]
                del self._cache[old_name]

            # Force the creation of the new file.
            #
            # fixme: I'm not sure that this is really the place for this.  We
            # do it because often the 'name' of the object is actually an
            # attribute of the object itself, and hence we want the serialized
            # state to reflect the new name... Hmmm...
            self._rebind(new_name, obj)

        # Update the name to filename map.
        del self._name_to_filename_map[old_name]
        self._name_to_filename_map[new_name] = new_filename

        # Move any attributes over to the new name.
        if old_name in self._attributes:
            self._attributes[new_name] = self._attributes[old_name]
            del self._attributes[old_name]
            self._save_attributes()

        return

    def _create_subcontext(self, name):
        """ Creates a sub-context of this context. """

        path = join(self.path, name)

        # Create a directory.
        os.mkdir(path)

        # Create a sub-context that represents the directory.
        sub = self._context_factory(name, path)

        # Update the name to filename map.
        self._name_to_filename_map[name] = name

        # Update the cache.
        self._cache[name] = sub

        return sub

    def _destroy_subcontext(self, name):
        """ Destroys a sub-context of this context. """

        return self._unbind(name)

    def _list_names(self):
        """ Lists the names bound in this context. """

        return list(self._name_to_filename_map.keys())

    # fixme: YFI this is not part of the protected 'Context' interface so
    # what is it doing here?
    def get_unique_name(self, name):

        ext = splitext(name)[1]

        # specially handle '.py' files
        if ext != '.py':
            return super(PyFSContext, self).get_unique_name(name)

        body = splitext(name)[0]
        names = self.list_names()
        i = 2
        unique = name
        while unique in names:
            unique = body + '_' + str(i) + '.py'
            i += 1

        return unique

    ###########################################################################
    # Protected 'DirContext' interface.
    ###########################################################################

    def _get_attributes(self, name):
        """ Returns the attributes of an object in this context. """

        attributes = self._attributes.setdefault(name, {})

        return attributes.copy()

    def _set_attributes(self, name, attributes):
        """ Sets the attributes of an object in this context. """

        self._attributes[name] = attributes
        self._save_attributes()

        return

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_filters(self):
        """ Returns the filters for this context. """

        return self.environment.get(self.FILTERS, [])

    def _get_object_serializers(self):
        """ Returns the object serializers for this context. """

        return self.environment.get(self.OBJECT_SERIALIZERS, [])

    def _context_factory(self, name, path):
        """ Create a sub-context. """

        return self.__class__(path=path, environment=self.environment)

    def _save_attributes(self):
        """ Saves all attributes to the attributes file. """

        path = join(self.path, self.ATTRIBUTES_FILE)

        f = open(path, 'wb')
        pickle.dump(self._attributes, f, 1)
        f.close()

        return

    #### Trait initializers ###################################################

    def __name_to_filename_map_default(self):
        """ Initializes the '_name_to_filename' trait. """

        # fixme: We should have a generalized filter mechanism (instead of
        # just 'glob' patterns we should have filter objects that can be a bit
        # more flexible in how they do the filtering).
        patterns = [join(self.path, filter) for filter in self._get_filters()]

        name_to_filename_map = {}
        for filename in os.listdir(self.path):
            path = join(self.path, filename)
            for pattern in patterns:
                if path in glob.glob(pattern):
                    break

            else:
                for serializer in self._get_object_serializers():
                    if serializer.can_load(filename):
                        # fixme: We should probably get the name from the
                        # serializer instead of assuming that we can just
                        # drop the file exension.
                        name, ext = os.path.splitext(filename)
                        break

                else:
                    name = filename

                name_to_filename_map[name] = filename

        return name_to_filename_map

    def __attributes_default(self):
        """ Initializes the '_attributes' trait. """

        attributes_file = File(join(self.path, self.ATTRIBUTES_FILE))
        if attributes_file.is_file:
            f = open(attributes_file.path, 'rb')
            attributes = pickle.load(f)
            f.close()

        else:
            attributes = {}

        return attributes

    #### Trait event handlers #################################################

    def _path_changed(self):
        """ Called when the context's path has changed. """

        basename = os.path.basename(self.path)

        self.name, ext = os.path.splitext(basename)

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" Naming exceptions. """


class NamingError(Exception):
    """ Base class for all naming exceptions.

    """

class InvalidNameError(NamingError):
    """ Invalid name.

    This exception is thrown when the name passed to a naming operation does
    not conform to the syntax of the naming system (or is empty etc).

    """

class NameAlreadyBoundError(NamingError):
    """ Name already bound.

    This exception is thrown when an attempt is made to bind a name that is
    already bound in the current context.

    """

class NameNotFoundError(NamingError):
    """ Name not found.

    This exception is thrown when a component of a name cannot be resolved
    because it is not bound in the current context.

    """

class NotContextError(NamingError):
    """ Not a context.

    This exception is thrown when a naming operation has reached a point where
    a context is required to continue the operation, but the resolved object
    is not a context.

    """

class OperationNotSupportedError(NamingError):
    """ The context does support the requested operation.

    """

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.action.api import Action
from traits.api import Unicode

# Local imports.
from apptools.appscripting.package_globals import get_script_manager


class StartRecordingAction(Action):
    """An action that starts the recording of changes to scriptable objects to
    a script."""

    #### 'Action' interface ###################################################

    name = Unicode("Start recording")

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Perform the action. """

        get_script_manager().start_recording()



"""Simple utility functions provided by the scripting API.
"""
# Author: Prabhu Ramachandran <prabhu [at] aero . iitb . ac . in>
# Copyright (c) 2008,  Prabhu Ramachandran
# License: BSD Style.

from .recorder import Recorder
from .recorder_with_ui import RecorderWithUI
from .package_globals import get_recorder, set_recorder


################################################################################
# Utility functions.
################################################################################
def start_recording(object, ui=True, **kw):
    """Convenience function to start recording.  Returns the recorder.

    Parameters:
    -----------

    object :  object to record.

    ui : bool specifying if a UI is to be shown or not

    kw : Keyword arguments to pass to the register function of the
    recorder.
    """
    if ui:
        r = RecorderWithUI(root=object)
        r.edit_traits(kind='live')
    else:
        r = Recorder()
    # Set the global recorder.
    set_recorder(r)
    r.recording = True
    r.register(object, **kw)
    return r

def stop_recording(object, save=True):
    """Stop recording the object.  If `save` is `True`, this will pop up
    a UI to ask where to save the script.
    """
    recorder = get_recorder()
    recorder.unregister(object)
    recorder.recording = False
    # Set the global recorder back to None
    set_recorder(None)
    # Save the script.
    if save:
        recorder.ui_save()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
#------------------------------------------------------------------------------
""" Abstract base class for a node in a preferences dialog. """


# Enthought library imports.
from traits.api import Delegate, Instance, Str

# Local imports.
from .i_preferences_page import IPreferencesPage
from .tree_item import TreeItem


class PreferencesNode(TreeItem):
    """ Abstract base class for a node in a preferences dialog.

    A preferences node has a name and an image which are used to represent the
    node in a preferences dialog (usually in the form of a tree).

    """

    #### 'PreferenceNode' interface ###########################################

    # The page's help identifier (optional). If a help Id *is* provided then
    # there will be a 'Help' button shown on the preference page.
    help_id = Delegate('page')

    # The page name (this is what is shown in the preferences dialog.
    name = Delegate('page')

    # The page that we are a node for.
    page = Instance(IPreferencesPage)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __str__(self):
        """ Returns the string representation of the item. """

        if self.page is None:
            s = 'root'

        else:
            s = self.page.name

        return s

    __repr__ = __str__

    ###########################################################################
    # 'PreferencesNode' interface.
    ###########################################################################

    def create_page(self, parent):
        """ Creates the preference page for this node. """

        return self.page.create_control(parent)

    def lookup(self, name):
        """ Returns the child of this node with the specified Id.

        Returns None if no such child exists.

        """

        for node in self.children:
            if node.name == name:
                break

        else:
            node = None

        return node

    ###########################################################################
    # Debugging interface.
    ###########################################################################

    def dump(self, indent=''):
        """ Pretty-print the node to stdout. """

        print(indent, 'Node', str(self))

        for child in self.children:
            child.dump(indent+'  ')

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from pyface.action.api import Action
from traits.api import Instance, Interface, List


class IPolicyManager(Interface):
    """The interface implemented by a policy manager.  A policy manager defines
    how permissions are assigned to users and stored.  A default policy manager
    is provided, but it may be replaced using the permissions manager."""

    # The list of PyFace policy management actions implemented by this policy.
    management_actions = List(Instance(Action))

    # The list of permissions assigned to the current user.
    user_permissions = List(Instance('apptools.permissions.api.Permission'))

    def bootstrapping(self):
        """Return True if the policy manager is bootstrapping.  Typically this
        is when no permissions have been assigned."""

    def load_user(self, user):
        """Load the policy for the given user.  user is an object that
        implements IUser.  If it is None then unload the policy for the current
        user."""

    def register_permission(self, permission):
        """Register the given permission defined by the application."""



""" The default implementation of a node in a preferences hierarchy. """



# Standard library imports.
import logging, threading

# Enthought library imports.
from traits.api import Any, Callable, Dict, HasTraits, Instance, List
from traits.api import Property, Str, Undefined, provides

# Local imports.
from .i_preferences import IPreferences


# Logging.
logger = logging.getLogger(__name__)


@provides(IPreferences)
class Preferences(HasTraits):
    """ The default implementation of a node in a preferences hierarchy. """


    #### 'IPreferences' interface #############################################

    # The absolute path to this node from the root node (the empty string if
    # this node *is* the root node).
    path = Property(Str)

    # The parent node (None if this node *is* the root node).
    parent = Instance(IPreferences)

    # The name of the node relative to its parent (the empty string if this
    # node *is* the root node).
    name = Str

    #### 'Preferences' interface ##############################################

    # The default name of the file used to persist the preferences (if no
    # filename is passed in to the 'load' and 'save' methods, then this is
    # used instead).
    filename = Str

    #### Protected 'Preferences' interface ####################################

    # A lock to make access to the node thread-safe.
    #
    # fixme: There *should* be no need to declare this as a trait, but if we
    # don't then we have problems using nodes in the preferences manager UI.
    # It is something to do with 'cloning' the node for use in a 'modal' traits
    # UI... Hmmm...
    _lk = Any

    # The node's children.
    _children = Dict(Str, IPreferences)

    # The node's preferences.
    _preferences = Dict(Str, Any)

    # Listeners for changes to the node's preferences.
    #
    # The callable must take 4 arguments, e.g::
    #
    # listener(node, key, old, new)
    _preferences_listeners = List(Callable)

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __init__(self, **traits):
        """ Constructor. """

        # A lock to make access to the '_children', '_preferences' and
        # '_preferences_listeners' traits thread-safe.
        self._lk = threading.Lock()

        # Base class constructor.
        super(Preferences, self).__init__(**traits)

        # If a filename has been specified then load the preferences from it.
        if len(self.filename) > 0:
            self.load()

        return

    ###########################################################################
    # 'IPreferences' interface.
    ###########################################################################

    #### Trait properties #####################################################

    def _get_path(self):
        """ Property getter. """

        names = []

        node = self
        while node.parent is not None:
            names.append(node.name)
            node = node.parent

        names.reverse()

        return '.'.join(names)

    #### Methods ##############################################################

    #### Methods where 'path' refers to a preference ####

    def get(self, path, default=None, inherit=False):
        """ Get the value of the preference at the specified path. """

        if len(path) == 0:
            raise ValueError('empty path')

        components = path.split('.')

        # If there is only one component in the path then the operation takes
        # place in this node.
        if len(components) == 1:
            value = self._get(path, Undefined)

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            node = self._get_child(components[0])
            if node is not None:
                value = node.get('.'.join(components[1:]), Undefined)

            else:
                value = Undefined

        # If inherited values are allowed then try those as well.
        #
        # e.g. 'acme.ui.widget.bgcolor'
        #      'acme.ui.bgcolor'
        #      'acme.bgcolor'
        #      'bgcolor'
        while inherit and value is Undefined and len(components) > 1:
            # Remove the penultimate component...
            #
            # e.g. 'acme.ui.widget.bgcolor' -> 'acme.ui.bgcolor'
            del components[-2]

            # ... and try that.
            value = self.get('.'.join(components), default=Undefined)

        if value is Undefined:
            value = default

        return value

    def remove(self, path):
        """ Remove the preference at the specified path. """

        if len(path) == 0:
            raise ValueError('empty path')

        components = path.split('.')

        # If there is only one component in the path then the operation takes
        # place in this node.
        if len(components) == 1:
            self._remove(path)

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            node = self._get_child(components[0])
            if node is not None:
                node.remove('.'.join(components[1:]))

        return

    def set(self, path, value):
        """ Set the value of the preference at the specified path. """

        if len(path) == 0:
            raise ValueError('empty path')

        components = path.split('.')

        # If there is only one component in the path then the operation takes
        # place in this node.
        if len(components) == 1:
            self._set(path, value)

        # Otherwise, find the next node (creating it if it doesn't exist)
        # and pass the rest of the path to that.
        else:
            node = self._node(components[0])
            node.set('.'.join(components[1:]), value)

        return

    #### Methods where 'path' refers to a node ####

    def clear(self, path=''):
        """ Remove all preferences from the node at the specified path. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            self._clear()

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._get_child(components[0])
            if node is not None:
                node.clear('.'.join(components[1:]))

        return

    def keys(self, path=''):
        """ Return the preference keys of the node at the specified path. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            keys = self._keys()

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._get_child(components[0])
            if node is not None:
                keys = node.keys('.'.join(components[1:]))

            else:
                keys = []

        return keys

    def node(self, path=''):
        """ Return the node at the specified path. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            node = self

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._node(components[0])
            node = node.node('.'.join(components[1:]))

        return node

    def node_exists(self, path=''):
        """ Return True if the node at the specified path exists. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            exists = True

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._get_child(components[0])
            if node is not None:
                exists = node.node_exists('.'.join(components[1:]))

            else:
                exists = False

        return exists

    def node_names(self, path=''):
        """ Return the names of the children of the node at the specified path.

        """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            names = self._node_names()

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._get_child(components[0])
            if node is not None:
                names = node.node_names('.'.join(components[1:]))

            else:
                names = []

        return names

    #### Persistence methods ####

    def flush(self):
        """ Force any changes in the node to the backing store.

        This includes any changes to the node's descendants.

        """

        self.save()

        return

    ###########################################################################
    # 'Preferences' interface.
    ###########################################################################

    #### Listener methods ####

    def add_preferences_listener(self, listener, path=''):
        """ Add a listener for changes to a node's preferences. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            names = self._add_preferences_listener(listener)

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._node(components[0])
            node.add_preferences_listener(listener, '.'.join(components[1:]))

        return

    def remove_preferences_listener(self, listener, path=''):
        """ Remove a listener for changes to a node's preferences. """

        # If the path is empty then the operation takes place in this node.
        if len(path) == 0:
            names = self._remove_preferences_listener(listener)

        # Otherwise, find the next node and pass the rest of the path to that.
        else:
            components = path.split('.')

            node = self._node(components[0])
            node.remove_preferences_listener(listener,'.'.join(components[1:]))

        return

    #### Persistence methods ####

    def load(self, file_or_filename=None):
        """ Load preferences from a file.

        This is a *merge* operation i.e. the contents of the file are added to
        the node.

        This implementation uses 'ConfigObj' files.

        """

        if file_or_filename is None:
            file_or_filename = self.filename

        logger.debug('loading preferences from <%s>', file_or_filename)

        # Do the import here so that we don't make 'ConfigObj' a requirement
        # if preferences aren't ever persisted (or a derived class chooses to
        # use a different persistence mechanism).
        from configobj import ConfigObj

        config_obj = ConfigObj(file_or_filename, encoding='utf-8')

        # 'name' is the section name, 'value' is a dictionary containing the
        # name/value pairs in the section (the actual preferences ;^).
        for name, value in list(config_obj.items()):
            # Create/get the node from the section name.
            components = name.split('.')

            node = self
            for component in components:
                node = node._node(component)

            # Add the contents of the section to the node.
            self._add_dictionary_to_node(node, value)

        return

    def save(self, file_or_filename=None):
        """ Save the node's preferences to a file.

        This implementation uses 'ConfigObj' files.

        """

        if file_or_filename is None:
            file_or_filename = self.filename

        # If no file or filename is specified then don't save the preferences!
        if len(file_or_filename) > 0:
            # Do the import here so that we don't make 'ConfigObj' a
            # requirement if preferences aren't ever persisted (or a derived
            # class chooses to use a different persistence mechanism).
            from configobj import ConfigObj

            logger.debug('saving preferences to <%s>', file_or_filename)

            config_obj = ConfigObj(file_or_filename, encoding='utf-8')
            self._add_node_to_dictionary(self, config_obj)
            config_obj.write()

        return

    ###########################################################################
    # Protected 'Preferences' interface.
    #
    # These are the only methods that should access the protected '_children'
    # and '_preferences' traits. This helps make it easy to subclass this class
    # to create other implementations (all the subclass has to do is to
    # implement these protected methods).
    #
    ###########################################################################

    def _add_dictionary_to_node(self, node, dictionary):
        """ Add the contents of a dictionary to a node's preferences. """

        self._lk.acquire()
        node._preferences.update(dictionary)
        self._lk.release()

        return

    def _add_node_to_dictionary(self, node, dictionary):
        """ Add a node's preferences to a dictionary. """

        # This method never manipulates the '_preferences' trait directly.
        # Instead it does eveything via the other protected methods and hence
        # doesn't need to grab the lock.
        if len(node._keys()) > 0:
            dictionary[node.path] = {}
            for key in node._keys():
                dictionary[node.path][key] = node._get(key)

        for name in node._node_names():
            self._add_node_to_dictionary(node._get_child(name), dictionary)

        return

    def _add_preferences_listener(self, listener):
        """ Add a listener for changes to thisnode's preferences. """

        self._lk.acquire()
        self._preferences_listeners.append(listener)
        self._lk.release()

        return

    def _clear(self):
        """ Remove all preferences from this node. """

        self._lk.acquire()
        self._preferences.clear()
        self._lk.release()

        return

    def _create_child(self, name):
        """ Create a child of this node with the specified name. """

        self._lk.acquire()
        child = self._children[name] = Preferences(name=name, parent=self)
        self._lk.release()

        return child

    def _get(self, key, default=None):
        """ Get the value of a preference in this node. """

        self._lk.acquire()
        value = self._preferences.get(key, default)
        self._lk.release()

        return value

    def _get_child(self, name):
        """ Return the child of this node with the specified name.

        Return None if no such child exists.

        """

        self._lk.acquire()
        child = self._children.get(name)
        self._lk.release()

        return child

    def _keys(self):
        """ Return the preference keys of this node. """

        self._lk.acquire()
        keys = list(self._preferences.keys())
        self._lk.release()

        return keys

    def _node(self, name):
        """ Return the child of this node with the specified name.

        Create the child node if it does not exist.

        """

        node = self._get_child(name)
        if node is None:
            node = self._create_child(name)

        return node

    def _node_names(self):
        """ Return the names of the children of this node. """

        self._lk.acquire()
        node_names = list(self._children.keys())
        self._lk.release()

        return node_names

    def _remove(self, name):
        """ Remove a preference value from this node. """

        self._lk.acquire()
        if name in self._preferences:
            del self._preferences[name]
        self._lk.release()

        return

    def _remove_preferences_listener(self, listener):
        """ Remove a listener for changes to the node's preferences. """

        self._lk.acquire()
        if listener in self._preferences_listeners:
            self._preferences_listeners.remove(listener)
        self._lk.release()

        return

    def _set(self, key, value):
        """ Set the value of a preference in this node. """
        
        # everything must be unicode encoded so that ConfigObj configuration
        # can properly serialize the data. Python str are supposed to be ASCII
        # encoded.
        value = str(value)

        self._lk.acquire()
        old = self._preferences.get(key)
        self._preferences[key] = value

        # If the value is unchanged then don't call the listeners!
        if old == value:
            listeners = []

        else:
            listeners = self._preferences_listeners[:]
        self._lk.release()

        for listener in listeners:
            listener(self, key, old, value)

        return

    ###########################################################################
    # Debugging interface.
    ###########################################################################

    def dump(self, indent=''):
        """ Dump the preferences hierarchy to stdout. """

        if indent == '':
            print()

        print(indent, 'Node(%s)' % self.name, self._preferences)
        indent += '  '

        for child in list(self._children.values()):
            child.dump(indent)

        return

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought application scripting package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import Any, Property, Undefined
from traits.traits import trait_cast

# Local imports.
from .package_globals import get_script_manager


# This is the guard that ensures that only outermost scriptable methods get
# recorded.
_outermost_call = True


def scriptable(func):
    """ This is the decorator applied to functions and methods to mark them as
    being scriptable.
    """

    def _scripter(*args, **kwargs):
        """ This is the wrapper that is returned in place of the scriptable
        method.
        """

        global _outermost_call

        if _outermost_call:
            _outermost_call = False

            # See if there is an script manager set.
            sm = get_script_manager()

            if func.__name__ == '__init__':
                sm.new_object(args[0], type(args[0]), args[1:], kwargs)

                try:
                    result = func(*args, **kwargs)
                finally:
                    _outermost_call = True
            else:
                # Record the ordinary method.
                try:
                    result = sm.record_method(func, args, kwargs)
                finally:
                    _outermost_call = True
        else:
            # We aren't at the outermost call so just invoke the method.
            result = func(*args, **kwargs)

        return result

    # Be a good citizen.
    _scripter.__name__ = func.__name__
    _scripter.__doc__ = func.__doc__
    _scripter.__dict__.update(func.__dict__)

    return _scripter


def _scriptable_get(obj, name):
    """ The getter for a scriptable trait. """

    global _outermost_call

    saved_outermost = _outermost_call
    _outermost_call = False

    try:
        result = getattr(obj, '_' + name, None)

        if result is None:
            result = obj.trait(name).default
    finally:
        _outermost_call = saved_outermost

    if saved_outermost:
        get_script_manager().record_trait_get(obj, name, result)

    return result


def _scriptable_set(obj, name, value):
    """ The setter for a scriptable trait. """

    if _outermost_call:
        get_script_manager().record_trait_set(obj, name, value)

    _name = '_' + name
    old_value = getattr(obj, _name, Undefined)

    if old_value is not value:
        setattr(obj, _name, value)
        obj.trait_property_changed(name, old_value, value)


def Scriptable(trait=Any, **metadata):
    """ Scriptable is a wrapper around another trait that makes it scriptable,
    ie. changes to its value can be recorded.  If a trait is read, but the
    value isn't set to another scriptable trait or passed to a scriptable
    method then the read will not be included in the recorded script.  To make
    sure a read is always recorded set the 'has_side_effects' argument to True.
    """

    trait = trait_cast(trait)
    metadata['default'] = trait.default_value()[1]

    return Property(_scriptable_get, _scriptable_set, trait=trait, **metadata)



""" Context adapter factory for Python instances. """


# Enthought library imports.
from apptools.naming.api import ContextAdapterFactory
from traits.api import List, Str
from apptools.type_manager import PythonObject

# Local imports.
from .instance_context_adapter import InstanceContextAdapter


class InstanceContextAdapterFactory(ContextAdapterFactory):
    """ Context adapter factoryfor Python instances. """

    #### 'ContextAdapterFactory' interface ####################################

    # The type of object that we adapt.
    adaptee_class = PythonObject

    #### 'InstanceContextAdapterFactory' interface ############################

    # By default every public attribute of an instance is exposed. Use the
    # following traits to either include or exclude attributes as appropriate.
    #
    # Regular expressions that describe the names of attributes to include.
    include = List(Str)

    # Regular expressions that describe the names of attributes to exclude.  By
    # default we exclude 'protected' and 'private' attributes and any
    # attributes that are artifacts of the traits mechanism.
    exclude = List(Str, ['_', 'trait_'])

    ###########################################################################
    # Protected 'AbstractAdapterFactory' interface.
    ###########################################################################

    def _adapt(self, adaptee, target_class, environment, context):
        """ Returns an adapter that adapts an object to the target class. """

        adapter = InstanceContextAdapter(
            adaptee     = adaptee,
            environment = environment,
            context     = context,
            include     = self.include,
            exclude     = self.exclude
        )

        return adapter

#### EOF ######################################################################



"""A special unpickler that gives you a state object and a special
pickler that lets you re-pickle that state.

The nice thing about creating state objects is that it does not import
any modules and does not create any instances.  Instead of instances
it creates `State` instances which have the same attributes as the
real object.  With this you can load a pickle (without even having the
modules on the machine), modify it and re-pickle it back.

NOTE: This module is not likely to work for very complex pickles but
 it should work for most common cases.

"""
# Author: Prabhu Ramachandran <prabhu@aero.iitb.ac.in>
# Copyright (c) 2006-2015, Prabhu Ramachandran
# License: BSD Style.

import sys
if sys.version_info[0] > 2:
    raise ImportError("This module does not work with Python 3.x")

import warnings
import pickle
import struct
from pickle import Pickler, Unpickler, dumps, BUILD, NEWOBJ, REDUCE, \
     MARK, OBJ, INST, BUILD, PicklingError, GLOBAL, \
     EXT1, EXT2, EXT4, _extension_registry, _keep_alive

from io import BytesIO


######################################################################
# `State` class
######################################################################
class State(dict):
    """Used to encapsulate the state of an instance in a very
    convenient form.  The '__METADATA__' attribute/key is a dictionary
    that has class specific details like the class name, module name
    etc.
    """
    def __init__(self, **kw):
        dict.__init__(self, **kw)
        self.__dict__ = self


######################################################################
# `StatePickler` class
######################################################################
class StatePickler(Pickler):
    """Pickles a `State` object back as a regular pickle that may be
    unpickled.
    """

    def __init__(self, file, protocol, bin=None):
        Pickler.__init__(self, file, protocol)
        self.bin = bin

    def save(self, obj):
        # Check the memo
        x = self.memo.get(id(obj))
        if x:
            self.write(self.get(x[0]))
            return

        if isinstance(obj, State):
            md = obj.__METADATA__
            typ = md['type']
            if typ == 'instance':
                self._state_instance(obj)
            elif typ in ['newobj', 'reduce']:
                self._state_reduce(obj)
            elif typ in ['class']:
                self._save_global(obj)
        else:
            Pickler.save(self, obj)

    def _state_instance(self, obj):
        md = obj.__METADATA__
        cls = md.get('class')
        cls_md = cls.__METADATA__

        memo  = self.memo
        write = self.write
        save  = self.save

        args = md.get('initargs')
        if len(args) > 0:
            _keep_alive(args, memo)

        write(MARK)

        if self.bin:
            save(cls)
            for arg in args:
                save(arg)
            write(OBJ)
        else:
            for arg in args:
                save(arg)
            write(INST + cls_md.get('module') + '\n' + cls_md.get('name') + '\n')

        self.memoize(obj)

        stuff = dict(obj.__dict__)
        stuff.pop('__METADATA__')

        if '__setstate_data__' in stuff:
            data = stuff.pop('__setstate_data__')
            _keep_alive(data, memo)
            save(data)
        else:
            save(stuff)
        write(BUILD)

    def _state_reduce(self, obj):
        # FIXME: this code is not as complete as pickle's reduce
        # handling code and is likely to not work in all cases.

        md = obj.__METADATA__
        func = md.get('class')
        func_md = func.__METADATA__
        args = md.get('initargs')
        state = dict(obj.__dict__)
        state.pop('__METADATA__')

        # This API is called by some subclasses

        # Assert that args is a tuple or None
        if not isinstance(args, tuple):
            if args is None:
                # A hack for Jim Fulton's ExtensionClass, now deprecated.
                # See load_reduce()
                warnings.warn("__basicnew__ special case is deprecated",
                              DeprecationWarning)
            else:
                raise PicklingError(
                    "args from reduce() should be a tuple")

        # Assert that func is callable
        #if not callable(func):
        #    raise PicklingError("func from reduce should be callable")

        save = self.save
        write = self.write

        # Protocol 2 special case: if func's name is __newobj__, use NEWOBJ
        if self.proto >= 2 and func_md.get("name", "") == "__newobj__":
            # FIXME: this is unlikely to work.
            cls = args[0]
            if not hasattr(cls, "__new__"):
                raise PicklingError(
                    "args[0] from __newobj__ args has no __new__")
            if obj is not None and cls is not obj.__class__:
                raise PicklingError(
                    "args[0] from __newobj__ args has the wrong class")
            args = args[1:]
            save(cls)
            save(args)
            write(NEWOBJ)
        else:
            save(func)
            save(args)
            write(REDUCE)

        if obj is not None:
            self.memoize(obj)

        if state is not None:
            if '__setstate_data__' in state:
                data = state.pop('__setstate_data__')
                save(data)
            else:
                save(state)
            write(BUILD)

    def _save_global(self, obj, name=None, pack=struct.pack):
        write = self.write
        memo = self.memo

        md = obj.__METADATA__

        if name is None:
            name = md.get('name')

        module = md.get('module')

        if self.proto >= 2:
            code = _extension_registry.get((module, name))
            if code:
                assert code > 0
                if code <= 0xff:
                    write(EXT1 + chr(code))
                elif code <= 0xffff:
                    write("%c%c%c" % (EXT2, code&0xff, code>>8))
                else:
                    write(EXT4 + pack("<i", code))
                return

        write(GLOBAL + module + '\n' + name + '\n')
        self.memoize(obj)


######################################################################
# `StateUnpickler` class
######################################################################
class StateUnpickler(Unpickler):
    def _instantiate(self, klass, k):
        args = tuple(self.stack[k+1:])
        del self.stack[k:]
        metadata = {'initargs': args, 'class': klass, 'type': 'instance'}
        value = State(__METADATA__=metadata)
        self.append(value)

    def load_build(self):
        stack = self.stack
        state = stack.pop()
        # Inst here is a State object.
        inst = stack[-1]
        slotstate = None
        if isinstance(state, tuple) and len(state) == 2:
            state, slotstate = state
        if state:
            if isinstance(state, tuple):
                inst.__dict__['__setstate_data__'] = state
            else:
                inst.__dict__.update(state)
        if slotstate:
            for k, v in list(slotstate.items()):
                setattr(inst, k, v)
    Unpickler.dispatch[BUILD] = load_build

    def load_newobj(self):
        args = self.stack.pop()
        cls = self.stack[-1]
        cls_md = cls.__METADATA__
        metadata = {'initargs': args, 'class': cls, 'type': 'newobj'}
        obj = State(__METADATA__ = metadata)
        #obj = cls.__new__(cls, *args)
        self.stack[-1] = obj
    Unpickler.dispatch[NEWOBJ] = load_newobj

    def load_reduce(self):
        stack = self.stack
        args = stack.pop()
        func = stack[-1]
        func_md = func.__METADATA__
        metadata = {'initargs': args, 'class': func, 'type': 'reduce'}
        value = State(__METADATA__ = metadata)
        stack[-1] = value
    Unpickler.dispatch[REDUCE] = load_reduce

    def load(self):
        # We overload the load_build method so update the dispatch dict.
        dispatch = self.dispatch
        dispatch[BUILD] = StateUnpickler.load_build
        dispatch[NEWOBJ] = StateUnpickler.load_newobj
        dispatch[REDUCE] = StateUnpickler.load_reduce
        # call the super class' method.
        ret = Unpickler.load(self)
        # Reset the Unpickler's dispatch
        dispatch[BUILD] = Unpickler.load_build
        dispatch[NEWOBJ] = Unpickler.load_newobj
        dispatch[REDUCE] = Unpickler.load_reduce
        return ret

    def find_class(self, module, name):
        metadata = {'module': module, 'name': name, 'type': 'class'}
        value = State(__METADATA__ = metadata)
        return value


######################################################################
# Utility functions.
######################################################################
def get_state(obj):
    """Return a State object given an object.  Useful for testing."""
    str = dumps(obj)
    return StateUnpickler(BytesIO(str)).load()

def dump_state(state, file, protocol=None, bin=None):
    """Dump the state (potentially modified) to given file."""
    StatePickler(file, protocol, bin).dump(state)

def dumps_state(state, protocol=None, bin=None):
    """Dump the state (potentially modified) to a string and return
    the string."""
    file = BytesIO()
    StatePickler(file, protocol, bin).dump(state)
    return file.getvalue()

def state2object(state):
    """Creates an object from a state."""
    s = dumps_state(state)
    return pickle.loads(s)

def load_state(file):
    """Loads the state from a file like object.  This does not import
    any modules."""
    return StateUnpickler(file).load()

def loads_state(string):
    """Loads the state from a string object.  This does not import any
    modules."""
    return StateUnpickler(BytesIO(string)).load()



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
""" The model for a tree view of a naming system. """


# Enthought library imports.
from apptools.naming.api import Binding
from pyface.tree.api import NodeTreeModel
from traits.api import Instance

# Local imports.
from .context_node_type import ContextNodeType
from .naming_node_manager import NamingNodeManager
from .object_node_type import ObjectNodeType


class NamingTreeModel(NodeTreeModel):
    """ The model for a tree view of a naming system. """

    #### 'TreeModel' interface ################################################

    # The root of the model.
    root = Instance(Binding)

    #### 'NodeTreeModel' interface ############################################

    # The node manager looks after all node types.
    node_manager = Instance(NamingNodeManager)

    ###########################################################################
    # 'NodeTreeModel' interface.
    ###########################################################################

    #### Trait initializers ###################################################

    def _node_manager_default(self):
        """ Initializes the node manaber trait. """

        node_manager = NamingNodeManager()
        node_manager.add_node_type(ContextNodeType())
        node_manager.add_node_type(ObjectNodeType())

        return node_manager

##### EOF #####################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought permissions package component>
#------------------------------------------------------------------------------


# Enthought library imports.
from traits.api import HasTraits, Instance, provides

# Local imports.
from .i_user_storage import IUserStorage, UserStorageError
from .persistent import Persistent, PersistentError


@provides(IUserStorage)
class UserStorage(HasTraits):
    """This implements a user database that pickles its data in a local file.
    """



    #### 'IUserStorage' interface #############################################

    capabilities = ['user_password', 'user_add', 'user_modify', 'user_delete']

    #### Private interface ####################################################

    # The persisted database.  The database itself is a dictionary, keyed by
    # the user name, and with a value that is a tuple of the description, blob
    # and clear text password.
    _db = Instance(Persistent)

    ###########################################################################
    # 'IUserStorage' interface.
    ###########################################################################

    def add_user(self, name, description, password):
        """Add a new user."""

        self._db.lock()

        try:
            users = self._db.read()

            if name in users:
                raise UserStorageError("The user \"%s\" already exists." % name)

            users[name] = (description, {}, password)
            self._db.write(users)
        finally:
            self._db.unlock()

    def authenticate_user(self, name, password):
        """Return the tuple of the user name, description, and blob if the user
        was successfully authenticated."""

        users = self._readonly_copy()

        try:
            description, blob, pword = users[name]
        except KeyError:
            raise UserStorageError("The name or password is invalid.")

        if password != pword:
            raise UserStorageError("The name or password is invalid.")

        return name, description, blob

    def delete_user(self, name):
        """Delete a user."""

        self._db.lock()

        try:
            users = self._db.read()

            try:
                del users[name]
            except KeyError:
                raise UserStorageError("The user \"%s\" does not exist." % name)

            self._db.write(users)
        finally:
            self._db.unlock()

    def is_empty(self):
        """See if the database is empty."""

        return (len(self._readonly_copy()) == 0)

    def matching_users(self, name):
        """Return the full name and description of all the users that match the
        given name."""

        # Return any user that starts with the name.
        users = [(full_name, description) for full_name, (description, _, _)
                in list(self._readonly_copy().items()) if full_name.startswith(name)]

        return sorted(users)

    def modify_user(self, name, description, password):
        """Update the description and password for the given user."""

        self._db.lock()

        try:
            users = self._db.read()

            try:
                _, blob, _ = users[name]
            except KeyError:
                raise UserStorageError("The user \"%s\" does not exist." % name)

            users[name] = (description, blob, password)
            self._db.write(users)
        finally:
            self._db.unlock()

    def update_blob(self, name, blob):
        """Update the blob for the given user."""

        self._db.lock()

        try:
            users = self._db.read()

            try:
                description, _, password = users[name]
            except KeyError:
                raise UserStorageError("The user \"%s\" does not exist." % name)

            users[name] = (description, blob, password)
            self._db.write(users)
        finally:
            self._db.unlock()

    def unauthenticate_user(self, user):
        """Unauthenticate the given user."""

        # There is nothing to do.
        return True

    def update_password(self, name, password):
        """Update the password for the given user."""

        self._db.lock()

        try:
            users = self._db.read()

            try:
                description, blob, _ = users[name]
            except KeyError:
                raise UserStorageError("The user \"%s\" does not exist." % name)

            users[name] = (description, blob, password)
            self._db.write(users)
        finally:
            self._db.unlock()

    ###########################################################################
    # Trait handlers.
    ###########################################################################

    def __db_default(self):
        """Return the default persisted database."""

        return Persistent(dict, 'ets_perms_userdb', "the user database")

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _readonly_copy(self):
        """Return the user database (which should not be modified)."""

        try:
            self._db.lock()

            try:
                data = self._db.read()
            finally:
                self._db.unlock()
        except PersistentError as e:
            raise UserStorageError(str(e))

        return data



""" The interface for a node in a preferences hierarchy. """


# Enthought library imports.
from traits.api import Instance, Interface, Str


class IPreferences(Interface):
    """ The interface for a node in a preferences hierarchy. """

    # The absolute path to this node from the root node (the empty string if
    # this node *is* the root node).
    path = Str

    # The parent node (None if this node *is* the root node).
    parent = Instance('IPreferences')

    # The name of the node relative to its parent (the empty string if this
    # node *is* the root node).
    name = Str

    #### Methods where 'path' refers to a preference ####

    def get(self, path, default=None, inherit=False):
        """ Get the value of the preference at the specified path.

        If no value exists for the path (or any part of the path does not
        exist) then return the default value.

        Preference values are *always* returned as strings.

        e.g::

          preferences.set('acme.ui.bgcolor', 'blue')
          preferences.get('acme.ui.bgcolor') -> 'blue'

          preferences.set('acme.ui.width', 100)
          preferences.get('acme.ui.width') -> '100'

          preferences.set('acme.ui.visible', True)
          preferences.get('acme.ui.visible') -> 'True'

        If 'inherit' is True then we allow 'inherited' preference values.

        e.g. If we are looking up::

          'acme.ui.widget.bgcolor'

        and it does not exist then we will also try::

          'acme.ui.bgcolor'
          'acme.bgcolor'
          'bgcolor'

        Raise a 'ValueError' exception if the path is the empty string.

        """

    def remove(self, path):
        """ Remove the preference at the specified path.

        Does nothing if no value exists for the path (or any part of the path
        does not exist.

        Raise a 'ValueError' exception if the path is the empty string.

        e.g.::

          preferences.remove('acme.ui.bgcolor')

        """

    def set(self, path, value):
        """ Set the value of the preference at the specified path.

        Any missing nodes are created automatically.

        Primitive Python types can be set, but preferences are *always*
        stored and returned as strings.

        e.g::

          preferences.set('acme.ui.bgcolor', 'blue')
          preferences.get('acme.ui.bgcolor') -> 'blue'

          preferences.set('acme.ui.width', 100)
          preferences.get('acme.ui.width') -> '100'

          preferences.set('acme.ui.visible', True)
          preferences.get('acme.ui.visible') -> 'True'

        Raise a 'ValueError' exception if the path is the empty string.

        """

    #### Methods where 'path' refers to a node ####

    def clear(self, path=''):
        """ Remove all preference from the node at the specified path.

        If the path is the empty string (the default) then remove the
        preferences in *this* node.

        This does not affect any of the node's children.

        e.g. To clear the preferences out of a node directly::

          preferences.clear()

        Or to clear the preferences of a node at a given path::

          preferences.clear('acme.ui')

        """

    def keys(self, path=''):
        """ Return the preference keys of the node at the specified path.

        If the path is the empty string (the default) then return the
        preference keys of *this* node.

        e.g::

          keys = preferences.keys('acme.ui')

        """

    def node(self, path=''):
        """ Return the node at the specified path.

        If the path is the empty string (the default) then return *this* node.

        Any missing nodes are created automatically.

        e.g::

          node = preferences.node('acme.ui')
          bgcolor = node.get('bgcolor')

        """

    def node_exists(self, path=''):
        """ Return True if the node at the specified path exists

        If the path is the empty string (the default) then return True.

        e.g::

          exists = preferences.exists('acme.ui')

        """

    def node_names(self, path=''):
        """ Return the names of the children of the node at the specified path.

        If the path is the empty string (the default) then return the names of
        the children of *this* node.

        e.g::

          names = preferences.node_names('acme.ui')

        """

    #### Persistence methods ####

    def flush(self):
        """ Force any changes in the node to the backing store.

        This includes any changes to the node's descendants.

        """

#### EOF ######################################################################



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought naming package component>
#------------------------------------------------------------------------------
from .exception import NamingError, InvalidNameError, NameAlreadyBoundError
from .exception import NameNotFoundError, NotContextError
from .exception import OperationNotSupportedError

from .address import Address
from .binding import Binding
from .context import Context
from .context_adapter import ContextAdapter
from .context_adapter_factory import ContextAdapterFactory
from .dynamic_context import DynamicContext
from .dir_context import DirContext
from .initial_context import InitialContext
from .initial_context_factory import InitialContextFactory
from .naming_event import NamingEvent
from .naming_manager import naming_manager
from .object_factory import ObjectFactory
from .object_serializer import ObjectSerializer
from .py_context import PyContext
from .py_object_factory import PyObjectFactory
from .pyfs_context import PyFSContext
from .pyfs_context_factory import PyFSContextFactory
from .pyfs_initial_context_factory import PyFSInitialContextFactory
from .pyfs_object_factory import PyFSObjectFactory
from .pyfs_state_factory import PyFSStateFactory
from .reference import Reference
from .referenceable import Referenceable
from .referenceable_state_factory import ReferenceableStateFactory
from .state_factory import StateFactory



""" An instance editor that allows total control over widget creation. """


# Enthought library imports.
from traits.etsconfig.api import ETSConfig
from traits.api import Any
from traitsui.api import EditorFactory

# fixme: We need to import the 'Editor' class from the appropriate toolkit.
exec('from traitsui.%s.editor import Editor' % ETSConfig.toolkit)


class _WidgetEditor(Editor):
    """ An instance editor that allows total control over widget creation. """

    #### '_WidgetEditor' interface ############################################

    # The toolkit-specific parent of the editor.
    parent = Any

    ###########################################################################
    # '_WidgetEditor' interface.
    ###########################################################################

    def init(self, parent):
        """ Initialize the editor. """

        self.parent = parent

        # fixme: What if there are no pages?!?
        page = self.object.pages[0]

        # Create the editor's control.
        self.control = page.create_control(parent)

        # Listen for the page being changed.
        self.object.on_trait_change(self._on_page_changed, 'selected_page')

        return

    def dispose(self):
        """ Dispose of the editor. """

        page = self.object.selected_page
        page.destroy_control()

        return

    def update_editor(self):
        """ Update the editor. """

        pass

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _on_page_changed(self, obj, trait_name, old, new):
        """ Dynamic trait change handler. """

        if old is not None:
            old.destroy_control()

        if new is not None:
            self.control = new.create_control(self.parent)

        return


class WidgetEditor(EditorFactory):
    """ A factory widget editors. """

    ###########################################################################
    # 'object' interface.
    ###########################################################################

    def __call__ (self, *args, **traits):
        """ Call the object. """

        return self.set(**traits)

    ###########################################################################
    # 'EditorFactory' interface.
    ###########################################################################

    def simple_editor(self, ui, object, name, description, parent):
        """ Create a simple editor. """

        editor = _WidgetEditor(
            parent,
            factory     = self,
            ui          = ui,
            object      = object,
            name        = name,
            description = description
        )

        return editor

    custom_editor   = simple_editor
    text_editor     = simple_editor
    readonly_editor = simple_editor

#### EOF ######################################################################



""" A page in a preferences dialog. """


# Enthought library imports.
from apptools.preferences.api import PreferencesHelper
from traits.api import Any, Dict, Str, provides

# Local imports.
from .i_preferences_page import IPreferencesPage


@provides(IPreferencesPage)
class PreferencesPage(PreferencesHelper):
    """ A page in a preferences dialog. """



    #### 'IPreferencesPage' interface #########################################

    # The page's category (e.g. 'General/Appearance'). The empty string means
    # that this is a top-level page.
    category = Str

    # DEPRECATED: The help_id was never fully implemented, and it's been
    # over two years (now 4/2009).  The original goal was for the the Help
    # button to automatically appear and connect to a help page with a
    # help_id.  Not removing the trait right now to avoid breaking code
    # that may be checking for this.
    #
    # Use PreferencesManager.show_help and trait show_help metadata instead.
    help_id = Str

    # The page name (this is what is shown in the preferences dialog.
    name = Str

    #### Private interface ####################################################

    # The traits UI that represents the page.
    _ui = Any

    # A dictionary containing the traits that have been changed since the
    # last call to 'apply'.
    _changed = Dict

    ###########################################################################
    # 'IPreferencesPage' interface.
    ###########################################################################

    def apply(self):
        """ Apply the page's preferences. """

        path = self._get_path()

        for trait_name, value in list(self._changed.items()):
            if self._is_preference_trait(trait_name):
                self.preferences.set('%s.%s' % (path, trait_name), value)

        self._changed.clear()

        return

    # fixme: We would like to be able to have the following API so that
    # developers are not forced into using traits UI for their preferences
    # pages, but at the moment I can't work out how to do it!
##     def create_control(self, parent):
##         """ Create the toolkit-specific control that represents the page. """

##         if self._ui is None:
##             self._ui = self.edit_traits(parent=parent, kind='subpanel')

##         return self._ui.control

##     def destroy_control(self):
##         """ Destroy the toolkit-specific control that represents the page. """

##         if self._ui is not None:
##             self._ui.dispose()
##             self._ui = None

##         return

    ###########################################################################
    # Private interface.
    ###########################################################################

    #### Trait change handlers ################################################

    def _anytrait_changed(self, trait_name, old, new):
        """ Static trait change handler.

        This is an important override! In the base-class when a trait is
        changed the preferences node is updated too. Here, we stop that from
        happening and just make a note of what changes have been made. The
        preferences node gets updated when the 'apply' method is called.

        """

        # If the trait was a list or dict '_items' trait then just treat it as
        # if the entire list or dict was changed.
        if trait_name.endswith('_items'):
            trait_name = trait_name[:-6]
            if self._is_preference_trait(trait_name):
                self._changed[trait_name] = getattr(self, trait_name)

        elif self._is_preference_trait(trait_name):
            self._changed[trait_name] = new

        return

    # fixme: Pretty much duplicated in 'PreferencesHelper' (except for the
    # class name of course!).
    def _is_preference_trait(self, trait_name):
        """ Return True if a trait represents a preference value. """

        if trait_name.startswith('_') or trait_name.endswith('_') \
           or trait_name in PreferencesPage.class_traits():
            return False

        return True

#### EOF ######################################################################
